@@ -3,7 +3,7 @@
   bignum.c -
 
   $Author: matz $
-  $Date: 2004/06/04 09:56:25 $
+  $Date: 2005/02/28 02:45:22 $
   created at: Fri Jun 10 00:48:55 JST 1994
 
   Copyright (C) 1993-2003 Yukihiro Matsumoto
@@ -747,6 +747,17 @@ big2ulong(x, type)
     return num;
 }
 
+unsigned long
+rb_big2ulong_pack(x)   
+    VALUE x;  
+{   
+    unsigned long num = big2ulong(x, "unsigned long");
+    if (!RBIGNUM(x)->sign) {
+	return -num;
+    }
+    return num;
+}  
+
 unsigned long
 rb_big2ulong(x)
     VALUE x;
@@ -1622,15 +1633,16 @@ rb_big_pow(x, y)
  */
 
 VALUE
-rb_big_and(x, y)
-    VALUE x, y;
+rb_big_and(xx, yy)
+    VALUE xx, yy;
 {
-    VALUE z;
+    volatile VALUE x, y, z;
     BDIGIT *ds1, *ds2, *zds;
     long i, l1, l2;
     char sign;
 
-    y = rb_to_int(y);
+    x = xx;
+    y = rb_to_int(yy);
     if (FIXNUM_P(y)) {
 	y = rb_int2big(FIX2LONG(y));
     }
@@ -1677,15 +1689,16 @@ rb_big_and(x, y)
  */
 
 VALUE
-rb_big_or(x, y)
-    VALUE x, y;
+rb_big_or(xx, yy)
+    VALUE xx, yy;
 {
-    VALUE z;
+    volatile VALUE x, y, z;
     BDIGIT *ds1, *ds2, *zds;
     long i, l1, l2;
     char sign;
 
-    y = rb_to_int(y);
+    x = xx;
+    y = rb_to_int(yy);
     if (FIXNUM_P(y)) {
 	y = rb_int2big(FIX2LONG(y));
     }
@@ -1734,15 +1747,17 @@ rb_big_or(x, y)
  */
 
 VALUE
-rb_big_xor(x, y)
-    VALUE x, y;
+rb_big_xor(xx, yy)
+    VALUE xx, yy;
 {
+    volatile VALUE x, y;
     VALUE z;
     BDIGIT *ds1, *ds2, *zds;
     long i, l1, l2;
     char sign;
 
-    y = rb_to_int(y);
+    x = xx;
+    y = rb_to_int(yy);
     if (FIXNUM_P(y)) {
 	y = rb_int2big(FIX2LONG(y));
     }
@@ -1956,7 +1971,7 @@ rb_big_coerce(x, y)
 	return rb_assoc_new(rb_int2big(FIX2LONG(y)), x);
     }
     else {
-	rb_raise(rb_eTypeError, "Can't coerce %s to Bignum",
+	rb_raise(rb_eTypeError, "can't coerce %s to Bignum",
 		 rb_obj_classname(y));
     }
     /* not reached */
@@ -1995,11 +2010,13 @@ rb_big_rand(max, rand_buf)
 	return rb_float_new(rand_buf[0]);
     }
     v = bignew(len,1);
+    len--;
+    BDIGITS(v)[len] = BDIGITS(max)[len] * rand_buf[len];    
     while (len--) {
 	BDIGITS(v)[len] = ((BDIGIT)~0) * rand_buf[len];
     }
 
-    return rb_big_modulo((VALUE)v, max);
+    return v;
 }
 
 /*
