@@ -2,8 +2,8 @@
 
   dir.c -
 
-  $Author: ocean $
-  $Date: 2004/04/12 01:45:27 $
+  $Author: eban $
+  $Date: 2005/04/11 13:26:32 $
   created at: Wed Jan  5 09:51:01 JST 1994
 
   Copyright (C) 1993-2003 Yukihiro Matsumoto
@@ -63,8 +63,8 @@ char *strchr _((char*,char));
 
 #include "util.h"
 
-#ifndef HAVE_LSTAT
-#define lstat(path,st) stat(path,st)
+#if !defined HAVE_LSTAT && !defined lstat
+#define lstat stat
 #endif
 
 #define FNM_NOESCAPE	0x01
@@ -93,8 +93,8 @@ char *strchr _((char*,char));
 
 static char *
 range(pat, test, flags)
-    char *pat;
-    char test;
+    const char *pat;
+    int test;
     int flags;
 {
     int not, ok = 0;
@@ -113,19 +113,19 @@ range(pat, test, flags)
 	    pat++;
 	cstart = cend = *pat++;
 	if (!cstart)
-	    return 0;
+	    return NULL;
 	if (*pat == '-' && pat[1] != ']') {
 	    pat++;
 	    if (escape && *pat == '\\')
 		pat++;
 	    cend = *pat++;
 	    if (!cend)
-		return 0;
+		return NULL;
 	}
 	if (downcase(cstart) <= test && test <= downcase(cend))
 	    ok = 1;
     }
-    return ok == not ? 0 : pat + 1;
+    return ok == not ? NULL : (char *)pat + 1;
 }
 
 #define ISDIRSEP(c) (pathname && isdirsep(c))
@@ -191,7 +191,7 @@ fnmatch(pat, string, flags)
 	    if (!*s || ISDIRSEP(*s) || PERIOD(s))
 		return FNM_NOMATCH;
 	    pat = range(pat, *s, flags);
-	    if (!pat)
+	    if (pat == NULL)
 		return FNM_NOMATCH;
 	    s++;
 	    break;
@@ -236,7 +236,11 @@ static void
 free_dir(dir)
     struct dir_data *dir;
 {
-    if (dir && dir->dir) closedir(dir->dir);
+    if (dir) {
+	if (dir->dir) closedir(dir->dir);
+	if (dir->path) free(dir->path);
+    }
+    free(dir);
 }
 
 static VALUE dir_close _((VALUE));
@@ -299,6 +303,7 @@ dir_initialize(dir, dirname)
  *  the block, and <code>Dir::open</code> returns the value of the
  *  block.
  */
+
 static VALUE
 dir_s_open(klass, dirname)
     VALUE klass, dirname;
@@ -459,10 +464,11 @@ dir_seek(dir, pos)
     VALUE dir, pos;
 {
     struct dir_data *dirp;
+    off_t p = NUM2OFFT(pos);
 
-#ifdef HAVE_SEEKDIR
     GetDIR(dir, dirp);
-    seekdir(dirp->dir, NUM2INT(pos));
+#ifdef HAVE_SEEKDIR
+    seekdir(dirp->dir, p);
     return dir;
 #else
     rb_notimplement();
@@ -538,24 +544,42 @@ dir_close(dir)
 
 static void
 dir_chdir(path)
-    const char *path;
+    VALUE path;
 {
-    if (chdir(path) < 0)
-	rb_sys_fail(path);
+    if (chdir(RSTRING(path)->ptr) < 0)
+	rb_sys_fail(RSTRING(path)->ptr);
 }
 
 static int chdir_blocking = 0;
 static VALUE chdir_thread = Qnil;
 
+struct chdir_data {
+    VALUE old_path, new_path;
+    int done;
+};
+
 static VALUE
-chdir_restore(path)
-    char *path;
+chdir_yield(args)
+    struct chdir_data *args;
 {
-    chdir_blocking--;
-    if (chdir_blocking == 0)
-	chdir_thread = Qnil;
-    dir_chdir(path);
-    free(path);
+    dir_chdir(args->new_path);
+    args->done = Qtrue;
+    chdir_blocking++;
+    if (chdir_thread == Qnil)
+	chdir_thread = rb_thread_current();
+    return rb_yield(args->new_path);
+}
+
+static VALUE
+chdir_restore(args)
+    struct chdir_data *args;
+{
+    if (args->done) {
+	chdir_blocking--;
+	if (chdir_blocking == 0)
+	    chdir_thread = Qnil;
+	dir_chdir(args->old_path);
+    }
     return Qnil;
 }
 
@@ -605,19 +629,18 @@ dir_s_chdir(argc, argv, obj)
     VALUE obj;
 {
     VALUE path = Qnil;
-    char *dist = "";
 
     rb_secure(2);
     if (rb_scan_args(argc, argv, "01", &path) == 1) {
 	SafeStringValue(path);
-	dist = RSTRING(path)->ptr;
     }
     else {
-	dist = getenv("HOME");
+	const char *dist = getenv("HOME");
 	if (!dist) {
 	    dist = getenv("LOGDIR");
 	    if (!dist) rb_raise(rb_eArgError, "HOME/LOGDIR not set");
 	}
+	path = rb_str_new2(dist);
     }
 
     if (chdir_blocking > 0) {
@@ -626,14 +649,15 @@ dir_s_chdir(argc, argv, obj)
     }
 
     if (rb_block_given_p()) {
+	struct chdir_data args;
 	char *cwd = my_getcwd();
-	chdir_blocking++;
-	if (chdir_thread == Qnil)
-	    chdir_thread = rb_thread_current();
-	dir_chdir(dist);
-	return rb_ensure(rb_yield, path, chdir_restore, (VALUE)cwd);
+
+	args.old_path = rb_tainted_str_new2(cwd); free(cwd);
+	args.new_path = path;
+	args.done = Qfalse;
+	return rb_ensure(chdir_yield, (VALUE)&args, chdir_restore, (VALUE)&args);
     }
-    dir_chdir(dist);
+    dir_chdir(path);
 
     return INT2FIX(0);
 }
@@ -735,13 +759,8 @@ dir_s_mkdir(argc, argv, obj)
     }
 
     check_dirname(&path);
-#ifndef _WIN32
     if (mkdir(RSTRING(path)->ptr, mode) == -1)
 	rb_sys_fail(RSTRING(path)->ptr);
-#else
-    if (mkdir(RSTRING(path)->ptr) == -1)
-	rb_sys_fail(RSTRING(path)->ptr);
-#endif
 
     return INT2FIX(0);
 }
@@ -858,8 +877,8 @@ remove_backslashes(p)
 #endif
 
 struct glob_args {
-    void (*func) _((const char*, VALUE));
-    const char *c;
+    void (*func) _((VALUE, VALUE));
+    VALUE c;
     VALUE v;
 };
 
@@ -870,14 +889,18 @@ glob_func_caller(val)
     VALUE val;
 {
     struct glob_args *args = (struct glob_args *)val;
-    (*args->func)(args->c, args->v);
+    VALUE path = args->c;
+
+    OBJ_TAINT(path);
+    RSTRING(path)->len = strlen(RSTRING(path)->ptr);
+    (*args->func)(path, args->v);
     return Qnil;
 }
 
 static int
 glob_call_func(func, path, arg)
-    void (*func) _((const char*, VALUE));
-    const char *path;
+    void (*func) _((VALUE, VALUE));
+    VALUE path;
     VALUE arg;
 {
     int status;
@@ -891,18 +914,22 @@ glob_call_func(func, path, arg)
     return status;
 }
 
+static int glob_helper(VALUE path, char *sub, int flags, void (*func)(VALUE,VALUE), VALUE arg);
+
 static int
-glob_helper(path, sub, flags, func, arg)
-    char *path;
+glob_helper(pv, sub, flags, func, arg)
+    VALUE pv;
     char *sub;
     int flags;
-    void (*func) _((const char*, VALUE));
+    void (*func) _((VALUE, VALUE));
     VALUE arg;
 {
     struct stat st;
-    char *p, *m;
+    char *p, *m, *path;
     int status = 0;
 
+    StringValue(pv);
+    path = RSTRING(pv)->ptr;
     p = sub ? sub : path;
     if (!has_magic(p, 0, flags)) {
 #if defined DOSISH
@@ -911,7 +938,7 @@ glob_helper(path, sub, flags, func, arg)
 	if (!(flags & FNM_NOESCAPE)) remove_backslashes(p);
 #endif
 	if (lstat(path, &st) == 0) {
-	    status = glob_call_func(func, path, arg);
+	    status = glob_call_func(func, pv, arg);
 	    if (status) return status;
 	}
 	else if (errno != ENOENT) {
@@ -926,7 +953,8 @@ glob_helper(path, sub, flags, func, arg)
 	if (*p == '/') p++;
 	m = strchr(p, '/');
 	if (has_magic(p, m, flags)) {
-	    char *dir, *base, *magic, *buf;
+	    char *dir, *base, *magic;
+	    VALUE buf;
 	    DIR *dirp;
 	    struct dirent *dp;
 	    int recursive = 0;
@@ -951,10 +979,9 @@ glob_helper(path, sub, flags, func, arg)
 		if (m && strcmp(magic, "**") == 0) {
 		    int n = strlen(base);
 		    recursive = 1;
-		    buf = ALLOC_N(char, n+strlen(m)+3);
-		    sprintf(buf, "%s%s", base, *base ? m : m+1);
-		    status = glob_helper(buf, buf+n, flags, func, arg);
-		    free(buf);
+		    buf = rb_str_new(0, n+strlen(m)+3);
+		    sprintf(RSTRING(buf)->ptr, "%s%s", base, *base ? m : m+1);
+		    status = glob_helper(buf, RSTRING(buf)->ptr+n, flags, func, arg);
 		    if (status) goto finalize;
 		}
 		dirp = opendir(dir);
@@ -983,36 +1010,32 @@ glob_helper(path, sub, flags, func, arg)
 			continue;
 		    if (fnmatch("*", dp->d_name, flags) != 0)
 			continue;
-		    buf = ALLOC_N(char, strlen(base)+NAMLEN(dp)+strlen(m)+6);
-		    sprintf(buf, "%s%s%s", base, (BASE) ? "/" : "", dp->d_name);
-		    if (lstat(buf, &st) < 0) {
-			if (errno != ENOENT) rb_sys_warning(buf);
-			free(buf);
+		    buf = rb_str_new(0, strlen(base)+NAMLEN(dp)+strlen(m)+6);
+		    sprintf(RSTRING(buf)->ptr, "%s%s%s", base, (BASE) ? "/" : "", dp->d_name);
+		    if (lstat(RSTRING(buf)->ptr, &st) < 0) {
+			if (errno != ENOENT) rb_sys_warning(RSTRING(buf)->ptr);
 			continue;
 		    }
 		    if (S_ISDIR(st.st_mode)) {
-			char *t = buf+strlen(buf);
+			char *t = RSTRING(buf)->ptr+strlen(RSTRING(buf)->ptr);
 		        strcpy(t, "/**");
 			strcpy(t+3, m);
 			status = glob_helper(buf, t, flags, func, arg);
-			free(buf);
 			if (status) break;
 			continue;
 		    }
-		    free(buf);
 		    continue;
 		}
 		if (fnmatch(magic, dp->d_name, flags) == 0) {
-		    buf = ALLOC_N(char, strlen(base)+NAMLEN(dp)+2);
-		    sprintf(buf, "%s%s%s", base, (BASE) ? "/" : "", dp->d_name);
+		    buf = rb_str_new(0, strlen(base)+NAMLEN(dp)+2);
+		    sprintf(RSTRING(buf)->ptr, "%s%s%s", base, (BASE) ? "/" : "", dp->d_name);
 		    if (!m) {
 			status = glob_call_func(func, buf, arg);
-			free(buf);
 			if (status) break;
 			continue;
 		    }
 		    tmp = ALLOC(struct d_link);
-		    tmp->path = buf;
+		    tmp->path = strdup(RSTRING(buf)->ptr);
 		    *tail = tmp;
 		    tail = &tmp->next;
 		}
@@ -1029,11 +1052,10 @@ glob_helper(path, sub, flags, func, arg)
 			    if (S_ISDIR(st.st_mode)) {
 				int len = strlen(link->path);
 				int mlen = strlen(m);
-				char *t = ALLOC_N(char, len+mlen+1);
 
-				sprintf(t, "%s%s", link->path, m);
-				status = glob_helper(t, t+len, flags, func, arg);
-				free(t);
+				buf = rb_str_new(0, len+mlen+1);
+				sprintf(RSTRING(buf)->ptr, "%s%s", link->path, m);
+				status = glob_helper(buf, RSTRING(buf)->ptr+len, flags, func, arg);
 			    }
 			}
 			else {
@@ -1055,28 +1077,42 @@ glob_helper(path, sub, flags, func, arg)
 
 static int
 rb_glob2(path, flags, func, arg)
-    const char *path;
+    VALUE path;
     int flags;
-    void (*func) _((const char*, VALUE));
+    void (*func) _((VALUE, VALUE));
     VALUE arg;
 {
-    char *buf;
     int status;
 
-    buf = ALLOC_N(char, strlen(path)+1);
-    strcpy(buf, path);
-    status = glob_helper(buf, 0, flags, func, arg);
-    free(buf);
+    status = glob_helper(path, 0, flags, func, arg);
     return status;
 }
 
+struct rb_glob_args {
+    void (*func) _((const char*, VALUE));
+    VALUE arg;
+};
+
+static void
+rb_glob_caller(path, a)
+    VALUE path, a;
+{
+    struct rb_glob_args *args = (struct rb_glob_args *)a;
+    (*args->func)(RSTRING(path)->ptr, args->arg);
+}
+
 void
 rb_glob(path, func, arg)
     const char *path;
     void (*func) _((const char*, VALUE));
     VALUE arg;
 {
-    int status = rb_glob2(path, 0, func, arg);
+    struct rb_glob_args args;
+    int status;
+
+    args.func = func;
+    args.arg = arg;
+    status = rb_glob2(rb_str_new2(path), 0, rb_glob_caller, (VALUE)&args);
 
     if (status) rb_jump_tag(status);
 }
@@ -1087,48 +1123,46 @@ rb_globi(path, func, arg)
     void (*func) _((const char*, VALUE));
     VALUE arg;
 {
-    int status = rb_glob2(path, FNM_CASEFOLD, func, arg);
+    struct rb_glob_args args;
+    int status;
+
+    args.func = func;
+    args.arg = arg;
+    status = rb_glob2(rb_str_new2(path), FNM_CASEFOLD, rb_glob_caller, (VALUE)&args);
 
     if (status) rb_jump_tag(status);
 }
 
 static void
 push_pattern(path, ary)
-    const char *path;
+    VALUE path;
     VALUE ary;
 {
-    VALUE str = rb_tainted_str_new2(path);
-
-    if (ary) {
-	rb_ary_push(ary, str);
-    }
-    else {
-	rb_yield(str);
-    }
+    rb_ary_push(ary, path);
 }
 
 static int
 push_globs(ary, s, flags)
     VALUE ary;
-    const char *s;
+    VALUE s;
     int flags;
 {
     return rb_glob2(s, flags, push_pattern, ary);
 }
 
 static int
-push_braces(ary, s, flags)
-    VALUE ary;
-    const char *s;
+push_braces(ary, str, flags)
+    VALUE ary, str;
     int flags;
 {
-    char *buf, *b;
-    const char *p, *t;
+    VALUE buf;
+    char *b;
+    const char *s, *p, *t;
     const char *lbrace, *rbrace;
     int nest = 0;
     int status = 0;
 
-    p = s;
+    s = p = RSTRING(str)->ptr;
     lbrace = rbrace = 0;
     while (*p) {
 	if (*p == '{') {
@@ -1148,9 +1182,6 @@ push_braces(ary, s, flags)
 
     if (lbrace && rbrace) {
 	int len = strlen(s);
-	buf = xmalloc(len + 1);
-	memcpy(buf, s, lbrace-s);
-	b = buf + (lbrace-s);
 	p = lbrace;
 	while (*p != '}') {
 	    t = p + 1;
@@ -1158,15 +1189,17 @@ push_braces(ary, s, flags)
 		/* skip inner braces */
 		if (*p == '{') while (*p!='}') p++;
 	    }
+	    buf = rb_str_new(0, len+1);
+	    memcpy(RSTRING(buf)->ptr, s, lbrace-s);
+	    b = RSTRING(buf)->ptr + (lbrace-s);
 	    memcpy(b, t, p-t);
 	    strcpy(b+(p-t), rbrace+1);
 	    status = push_braces(ary, buf, flags);
 	    if (status) break;
 	}
-	free(buf);
     }
     else {
-	status = push_globs(ary, s, flags);
+	status = push_globs(ary, str, flags);
     }
 
     return status;
@@ -1180,26 +1213,21 @@ rb_push_glob(str, flags)
     int flags;
 {
     const char *p, *pend;
-    char *buf;
+    volatile VALUE buf;
     char *t;
     int nest, maxnest;
     int status = 0;
     int noescape = flags & FNM_NOESCAPE;
     VALUE ary;
 
-    if (rb_block_given_p())
-	ary = 0;
-    else
-	ary = rb_ary_new();
-
+    ary = rb_ary_new();
     SafeStringValue(str);
-    buf = xmalloc(RSTRING(str)->len + 1);
-
     p = RSTRING(str)->ptr;
     pend = p + RSTRING(str)->len;
 
     while (p < pend) {
-	t = buf;
+	buf = rb_str_new(0, pend - p + 1);
+	t = RSTRING(buf)->ptr;
 	nest = maxnest = 0;
 	while (p < pend && isdelim(*p)) p++;
 	while (p < pend && !isdelim(*p)) {
@@ -1222,10 +1250,11 @@ rb_push_glob(str, flags)
 	}
 	/* else unmatched braces */
     }
-    free(buf);
-
     if (status) rb_jump_tag(status);
-
+    if (rb_block_given_p()) {
+	rb_ary_each(ary);
+	return Qnil;
+    }
     return ary;
 }
 
@@ -1247,7 +1276,7 @@ dir_s_aref(obj, str)
 /*
  *  call-seq:
  *     Dir.glob( string, [flags] ) => array
- *     Dir.glob( string, [flags] ) {| filename | block }  => false
+ *     Dir.glob( string, [flags] ) {| filename | block }  => nil
  *
  *  Returns the filenames found by expanding the pattern given in
  *  <i>string</i>, either as an <i>array</i> or as parameters to the
@@ -1282,6 +1311,22 @@ dir_s_glob(argc, argv, obj)
     return rb_push_glob(str, flags);
 }
 
+static VALUE
+dir_open_dir(path)
+    VALUE path;
+{
+    struct dir_data *dp;
+    VALUE dir = rb_funcall(rb_cDir, rb_intern("open"), 1, path);
+
+    if (TYPE(dir) != T_DATA ||
+	RDATA(dir)->dfree != (RUBY_DATA_FUNC)free_dir) {
+	rb_raise(rb_eTypeError, "wrong argument type %s (expected Dir)",
+		 rb_obj_classname(dir));
+    }
+    return dir;
+}
+
+
 /*
  *  call-seq:
  *     Dir.foreach( dirname ) {| filename | block }  => nil
@@ -1305,7 +1350,7 @@ dir_foreach(io, dirname)
 {
     VALUE dir;
 
-    dir = rb_funcall(rb_cDir, rb_intern("open"), 1, dirname);
+    dir = dir_open_dir(dirname);
     rb_ensure(dir_each, dir, dir_close, dir);
     return Qnil;
 }
@@ -1327,7 +1372,7 @@ dir_entries(io, dirname)
 {
     VALUE dir;
 
-    dir = rb_funcall(rb_cDir, rb_intern("open"), 1, dirname);
+    dir = dir_open_dir(dirname);
     return rb_ensure(rb_Array, dir, dir_close, dir);
 }
 
