@@ -2,8 +2,8 @@
 
   error.c -
 
-  $Author: nobu $
-  $Date: 2004/07/16 02:17:59 $
+  $Author: eban $
+  $Date: 2005/03/18 03:15:48 $
   created at: Mon Aug  9 16:11:34 JST 1993
 
   Copyright (C) 1993-2003 Yukihiro Matsumoto
@@ -353,10 +353,7 @@ exc_initialize(argc, argv, exc)
 {
     VALUE arg;
 
-    if (rb_scan_args(argc, argv, "01", &arg) == 1) {
-	VALUE mesg = arg;
-	StringValue(mesg);	/* ensure mesg can be converted to String */
-    }
+    rb_scan_args(argc, argv, "01", &arg);
     rb_iv_set(exc, "mesg", arg);
     rb_iv_set(exc, "bt", Qnil);
 
@@ -681,7 +678,7 @@ name_err_to_s(exc)
  * call-seq:
  *   NoMethodError.new(msg, name [, args])  => no_method_error
  *
- * Contruct a NoMethodError exception for a method of the given name
+ * Construct a NoMethodError exception for a method of the given name
  * called with the given arguments. The name may be accessed using
  * the <code>#name</code> method on the resulting object, and the
  * arguments using the <code>#args</code> method.
@@ -911,7 +908,7 @@ syserr_initialize(argc, argv, self)
     if (!NIL_P(mesg)) {
 	VALUE str = mesg;
 	StringValue(str);
-	mesg = rb_str_new(0, strlen(err)+RSTRING(mesg)->len+3);
+	mesg = rb_str_new(0, strlen(err)+RSTRING(str)->len+3);
 	sprintf(RSTRING(mesg)->ptr, "%s - %.*s", err,
 		(int)RSTRING(str)->len, RSTRING(str)->ptr);
 	rb_str_resize(mesg, strlen(RSTRING(mesg)->ptr));
@@ -1083,7 +1080,7 @@ void
 rb_notimplement()
 {
     rb_raise(rb_eNotImpError,
-	     "The %s() function is unimplemented on this machine",
+	     "the %s() function is unimplemented on this machine",
 	     rb_id2name(ruby_frame->last_func));
 }
 
@@ -1152,14 +1149,14 @@ rb_sys_warning(fmt, va_alist)
 
 void
 rb_load_fail(path)
-    char *path;
+    const char *path;
 {
     rb_loaderror("%s -- %s", strerror(errno), path);
 }
 
 void
 rb_error_frozen(what)
-    char *what;
+    const char *what;
 {
     rb_raise(rb_eTypeError, "can't modify frozen %s", what);
 }
