@@ -19,6 +19,7 @@ static VALUE mReadline;
 
 #define COMPLETION_PROC "completion_proc"
 #define COMPLETION_CASE_FOLD "completion_case_fold"
+static ID completion_proc, completion_case_fold;
 
 #ifndef READLINE_42_OR_LATER
 # define rl_filename_completion_function filename_completion_function
@@ -45,8 +46,10 @@ readline_readline(argc, argv, self)
     char *buff;
     int status;
 
+    rb_secure(4);
     if (rb_scan_args(argc, argv, "02", &tmp, &add_hist) > 0) {
-	prompt = StringValuePtr(tmp);
+	SafeStringValue(tmp);
+	prompt = RSTRING(tmp)->ptr;
     }
 
     if (!isatty(0) && errno == EBADF) rb_raise(rb_eIOError, "stdin closed");
@@ -82,16 +85,18 @@ readline_s_set_completion_proc(self, proc)
     VALUE self;
     VALUE proc;
 {
+    rb_secure(4);
     if (!rb_respond_to(proc, rb_intern("call")))
 	rb_raise(rb_eArgError, "argument must respond to `call'");
-    return rb_iv_set(mReadline, COMPLETION_PROC, proc);
+    return rb_ivar_set(mReadline, completion_proc, proc);
 }
 
 static VALUE
 readline_s_get_completion_proc(self)
     VALUE self;
 {
-    return rb_iv_get(mReadline, COMPLETION_PROC);
+    rb_secure(4);
+    return rb_attr_get(mReadline, completion_proc);
 }
 
 static VALUE
@@ -99,14 +104,16 @@ readline_s_set_completion_case_fold(self, val)
     VALUE self;
     VALUE val;
 {
-    return rb_iv_set(mReadline, COMPLETION_CASE_FOLD, val);
+    rb_secure(4);
+    return rb_ivar_set(mReadline, completion_case_fold, val);
 }
 
 static VALUE
 readline_s_get_completion_case_fold(self)
     VALUE self;
 {
-    return rb_iv_get(mReadline, COMPLETION_CASE_FOLD);
+    rb_secure(4);
+    return rb_attr_get(mReadline, completion_case_fold);
 }
 
 static char **
@@ -120,11 +127,11 @@ readline_attempted_completion_function(text, start, end)
     int case_fold;
     int i, matches;
 
-    proc = rb_iv_get(mReadline, COMPLETION_PROC);
+    proc = rb_attr_get(mReadline, completion_proc);
     if (NIL_P(proc))
 	return NULL;
     rl_attempted_completion_over = 1;
-    case_fold = RTEST(rb_iv_get(mReadline, COMPLETION_CASE_FOLD));
+    case_fold = RTEST(rb_attr_get(mReadline, completion_case_fold));
     ary = rb_funcall(proc, rb_intern("call"), 1, rb_tainted_str_new2(text));
     if (TYPE(ary) != T_ARRAY)
 	ary = rb_Array(ary);
@@ -178,6 +185,7 @@ static VALUE
 readline_s_vi_editing_mode(self)
     VALUE self;
 {
+    rb_secure(4);
     rl_vi_editing_mode(1,0);
     return Qnil;
 }
@@ -186,6 +194,7 @@ static VALUE
 readline_s_emacs_editing_mode(self)
     VALUE self;
 {
+    rb_secure(4);
     rl_emacs_editing_mode(1,0);
     return Qnil;
 }
@@ -195,10 +204,17 @@ readline_s_set_completion_append_character(self, str)
     VALUE self, str;
 {
 #ifdef READLINE_21_OR_LATER
-    if (NIL_P(str) || !StringValuePtr(str) || !RSTRING(str)->len) {
+    rb_secure(4);
+    if (NIL_P(str)) {
 	rl_completion_append_character = '\0';
-    } else {
-	rl_completion_append_character = RSTRING(str)->ptr[0];
+    }
+    else {
+	SafeStringValue(str);
+	if (RSTRING(str)->len == 0) {
+	    rl_completion_append_character = '\0';
+	} else {
+	    rl_completion_append_character = RSTRING(str)->ptr[0];
+	}
     }
 
     return self;
@@ -214,6 +230,7 @@ readline_s_get_completion_append_character(self)
 #ifdef READLINE_21_OR_LATER
     VALUE str;
 
+    rb_secure(4);
     if (rl_completion_append_character == '\0')
 	return Qnil;
 
@@ -233,7 +250,8 @@ readline_s_set_basic_word_break_characters(self, str)
 #ifdef READLINE_21_OR_LATER
     static char *basic_word_break_characters = NULL;
 
-    StringValue(str);
+    rb_secure(4);
+    SafeStringValue(str);
     if (basic_word_break_characters == NULL) {
 	basic_word_break_characters =
 	    ALLOC_N(char, RSTRING(str)->len + 1);
@@ -257,9 +275,10 @@ readline_s_get_basic_word_break_characters(self, str)
     VALUE self, str;
 {
 #ifdef READLINE_21_OR_LATER
+    rb_secure(4);
     if (rl_basic_word_break_characters == NULL)
 	return Qnil;
-    return rb_str_new2(rl_basic_word_break_characters);
+    return rb_tainted_str_new2(rl_basic_word_break_characters);
 #else
     rb_notimplement();
 #endif /* READLINE_21_OR_LATER */
@@ -272,7 +291,8 @@ readline_s_set_completer_word_break_characters(self, str)
 #ifdef READLINE_21_OR_LATER
     static char *completer_word_break_characters = NULL;
 
-    StringValue(str);
+    rb_secure(4);
+    SafeStringValue(str);
     if (completer_word_break_characters == NULL) {
 	completer_word_break_characters =
 	    ALLOC_N(char, RSTRING(str)->len + 1);
@@ -296,9 +316,10 @@ readline_s_get_completer_word_break_characters(self, str)
     VALUE self, str;
 {
 #ifdef READLINE_21_OR_LATER
+    rb_secure(4);
     if (rl_completer_word_break_characters == NULL)
 	return Qnil;
-    return rb_str_new2(rl_completer_word_break_characters);
+    return rb_tainted_str_new2(rl_completer_word_break_characters);
 #else
     rb_notimplement();
 #endif /* READLINE_21_OR_LATER */
@@ -311,7 +332,8 @@ readline_s_set_basic_quote_characters(self, str)
 #ifdef READLINE_21_OR_LATER
     static char *basic_quote_characters = NULL;
 
-    StringValue(str);
+    rb_secure(4);
+    SafeStringValue(str);
     if (basic_quote_characters == NULL) {
 	basic_quote_characters =
 	    ALLOC_N(char, RSTRING(str)->len + 1);
@@ -335,9 +357,10 @@ readline_s_get_basic_quote_characters(self, str)
     VALUE self, str;
 {
 #ifdef READLINE_21_OR_LATER
+    rb_secure(4);
     if (rl_basic_quote_characters == NULL)
 	return Qnil;
-    return rb_str_new2(rl_basic_quote_characters);
+    return rb_tainted_str_new2(rl_basic_quote_characters);
 #else
     rb_notimplement();
 #endif /* READLINE_21_OR_LATER */
@@ -350,7 +373,8 @@ readline_s_set_completer_quote_characters(self, str)
 #ifdef READLINE_21_OR_LATER
     static char *completer_quote_characters = NULL;
 
-    StringValue(str);
+    rb_secure(4);
+    SafeStringValue(str);
     if (completer_quote_characters == NULL) {
 	completer_quote_characters =
 	    ALLOC_N(char, RSTRING(str)->len + 1);
@@ -374,9 +398,10 @@ readline_s_get_completer_quote_characters(self, str)
     VALUE self, str;
 {
 #ifdef READLINE_21_OR_LATER
+    rb_secure(4);
     if (rl_completer_quote_characters == NULL)
 	return Qnil;
-    return rb_str_new2(rl_completer_quote_characters);
+    return rb_tainted_str_new2(rl_completer_quote_characters);
 #else
     rb_notimplement();
 #endif /* READLINE_21_OR_LATER */
@@ -389,7 +414,8 @@ readline_s_set_filename_quote_characters(self, str)
 #ifdef READLINE_21_OR_LATER
     static char *filename_quote_characters = NULL;
 
-    StringValue(str);
+    rb_secure(4);
+    SafeStringValue(str);
     if (filename_quote_characters == NULL) {
 	filename_quote_characters =
 	    ALLOC_N(char, RSTRING(str)->len + 1);
@@ -413,9 +439,10 @@ readline_s_get_filename_quote_characters(self, str)
     VALUE self, str;
 {
 #ifdef READLINE_21_OR_LATER
+    rb_secure(4);
     if (rl_filename_quote_characters == NULL)
 	return Qnil;
-    return rb_str_new2(rl_filename_quote_characters);
+    return rb_tainted_str_new2(rl_filename_quote_characters);
 #else
     rb_notimplement();
 #endif /* READLINE_21_OR_LATER */
@@ -428,6 +455,7 @@ rb_remove_history(index)
     HIST_ENTRY *entry;
     VALUE val;
 
+    rb_secure(4);
     entry = remove_history(index);
     if (entry) {
         val = rb_tainted_str_new2(entry->line);
@@ -453,13 +481,14 @@ hist_get(self, index)
     HISTORY_STATE *state;
     int i;
 
+    rb_secure(4);
     state = history_get_history_state();
     i = NUM2INT(index);
     if (i < 0) {
         i += state->length;
     }
     if (i < 0 || i > state->length - 1) {
-	rb_raise(rb_eIndexError, "Invalid index");
+	rb_raise(rb_eIndexError, "invalid index");
     }
     return rb_tainted_str_new2(state->entries[i]->line);
 }
@@ -474,15 +503,17 @@ hist_set(self, index, str)
     VALUE s = str;
     int i;
 
+    rb_secure(4);
     state = history_get_history_state();
     i = NUM2INT(index);
     if (i < 0) {
         i += state->length;
     }
     if (i < 0 || i > state->length - 1) {
-	rb_raise(rb_eIndexError, "Invalid index");
+	rb_raise(rb_eIndexError, "invalid index");
     }
-    replace_history_entry(i, StringValuePtr(s), NULL);
+    SafeStringValue(str);
+    replace_history_entry(i, RSTRING(str)->ptr, NULL);
     return str;
 }
 
@@ -491,7 +522,9 @@ hist_push(self, str)
     VALUE self;
     VALUE str;
 {
-    add_history(StringValuePtr(str));
+    rb_secure(4);
+    SafeStringValue(str);
+    add_history(RSTRING(str)->ptr);
     return self;
 }
 
@@ -503,9 +536,11 @@ hist_push_method(argc, argv, self)
 {
     VALUE str;
     
+    rb_secure(4);
     while (argc--) {
 	str = *argv++;
-	add_history(StringValuePtr(str));
+	SafeStringValue(str);
+	add_history(RSTRING(str)->ptr);
     }
     return self;
 }
@@ -516,6 +551,7 @@ hist_pop(self)
 {
     HISTORY_STATE *state;
 
+    rb_secure(4);
     state = history_get_history_state();
     if (state->length > 0) {
 	return rb_remove_history(state->length - 1);
@@ -530,6 +566,7 @@ hist_shift(self)
 {
     HISTORY_STATE *state;
 
+    rb_secure(4);
     state = history_get_history_state();
     if (state->length > 0) {
 	return rb_remove_history(0);
@@ -545,6 +582,7 @@ hist_each(self)
     HISTORY_STATE *state;
     int i;
 
+    rb_secure(4);
     state = history_get_history_state();
     for (i = 0; i < state->length; i++) {
 	rb_yield(rb_tainted_str_new2(state->entries[i]->line));
@@ -558,6 +596,7 @@ hist_length(self)
 {
     HISTORY_STATE *state;
 
+    rb_secure(4);
     state = history_get_history_state();
     return INT2NUM(state->length);
 }
@@ -568,6 +607,7 @@ hist_empty_p(self)
 {
     HISTORY_STATE *state;
 
+    rb_secure(4);
     state = history_get_history_state();
     if (state->length == 0)
 	return Qtrue;
@@ -583,12 +623,13 @@ hist_delete_at(self, index)
     HISTORY_STATE *state;
     int i;
 
+    rb_secure(4);
     state = history_get_history_state();
     i = NUM2INT(index);
     if (i < 0)
         i += state->length;
     if (i < 0 || i > state->length - 1) {
-	rb_raise(rb_eIndexError, "Invalid index");
+	rb_raise(rb_eIndexError, "invalid index");
     }
     return rb_remove_history(i);
 }
@@ -657,6 +698,9 @@ Init_readline()
 
     using_history();
 
+    completion_proc = rb_intern(COMPLETION_PROC);
+    completion_case_fold = rb_intern(COMPLETION_CASE_FOLD);
+
     mReadline = rb_define_module("Readline");
     rb_define_module_function(mReadline, "readline",
 			      readline_readline, -1);
