@@ -2,8 +2,8 @@
 
   stringio.c -
 
-  $Author: matz $
-  $Date: 2004/04/05 07:45:23 $
+  $Author: nobu $
+  $Date: 2005/04/30 04:11:33 $
   $RoughId: stringio.c,v 1.13 2002/03/14 03:24:18 nobu Exp $
   created at: Tue Feb 19 04:10:38 JST 2002
 
@@ -14,9 +14,13 @@
 
 #include "ruby.h"
 #include "rubyio.h"
+#if defined(HAVE_FCNTL_H) || defined(_WIN32)
+#include <fcntl.h>
+#elif defined(HAVE_SYS_FCNTL_H)
+#include <sys/fcntl.h>
+#endif
 
-#define STRIO_APPEND 4
-#define STRIO_EOF 8
+#define STRIO_EOF FMODE_SYNC
 
 struct StringIO {
     VALUE string;
@@ -74,7 +78,7 @@ check_strio(self)
 {
     Check_Type(self, T_DATA);
     if (!IS_STRIO(self)) {
-	rb_raise(rb_eTypeError, "wrong argument type %s (expected String::IO)",
+	rb_raise(rb_eTypeError, "wrong argument type %s (expected StringIO)",
 		 rb_class2name(CLASS_OF(self)));
     }
     return DATA_PTR(self);
@@ -183,6 +187,13 @@ strio_s_allocate(klass)
     return Data_Wrap_Struct(klass, strio_mark, strio_free, 0);
 }
 
+/*
+ * call-seq: StringIO.open(string=""[, mode]) {|strio| ...}
+ *
+ * Equivalent to StringIO.new except that when it is called with a block, it
+ * yields with the new instance and closes it, and returns the result which
+ * returned from the block.
+ */
 static VALUE
 strio_s_open(argc, argv, klass)
     int argc;
@@ -194,6 +205,11 @@ strio_s_open(argc, argv, klass)
     return rb_ensure(rb_yield, obj, strio_finalize, obj);
 }
 
+/*
+ * call-seq: StringIO.new(string=""[, mode])
+ *
+ * Creates new StringIO instance from with _string_ and _mode_.
+ */
 static VALUE
 strio_initialize(argc, argv, self)
     int argc;
@@ -202,7 +218,7 @@ strio_initialize(argc, argv, self)
 {
     struct StringIO *ptr = check_strio(self);
     VALUE string, mode;
-    const char* m;
+    int trunc = Qfalse;
 
     if (!ptr) {
 	DATA_PTR(self) = ptr = strio_alloc();
@@ -210,21 +226,23 @@ strio_initialize(argc, argv, self)
     rb_call_super(0, 0);
     switch (rb_scan_args(argc, argv, "02", &string, &mode)) {
       case 2:
-	StringValue(mode);
+	if (FIXNUM_P(mode)) {
+	    int flags = FIX2INT(mode);
+	    ptr->flags = rb_io_modenum_flags(flags);
+	    trunc = flags & O_TRUNC;
+	}
+	else {
+	    const char *m = StringValueCStr(mode);
+	    ptr->flags = rb_io_mode_flags(m);
+	    trunc = *m == 'w';
+	}
 	StringValue(string);
-	if (!(m = RSTRING(mode)->ptr)) m = "";
-	ptr->flags = rb_io_mode_flags(m);
 	if ((ptr->flags & FMODE_WRITABLE) && OBJ_FROZEN(string)) {
 	    errno = EACCES;
 	    rb_sys_fail(0);
 	}
-	switch (*m) {
-	  case 'a':
-	    ptr->flags |= STRIO_APPEND;
-	    break;
-	  case 'w':
+	if (trunc) {
 	    rb_str_resize(string, 0);
-	    break;
 	}
 	break;
       case 1:
@@ -250,6 +268,9 @@ strio_finalize(self)
     return self;
 }
 
+/*
+ * Returns +false+.  Just for compatibility to IO.
+ */
 static VALUE
 strio_false(self)
     VALUE self;
@@ -258,6 +279,9 @@ strio_false(self)
     return Qfalse;
 }
 
+/*
+ * Returns +nil+.  Just for compatibility to IO.
+ */
 static VALUE
 strio_nil(self)
     VALUE self;
@@ -266,6 +290,9 @@ strio_nil(self)
     return Qnil;
 }
 
+/*
+ * Returns *strio* itself.  Just for compatibility to IO.
+ */
 static VALUE
 strio_self(self)
     VALUE self;
@@ -274,6 +301,9 @@ strio_self(self)
     return self;
 }
 
+/*
+ * Returns 0.  Just for compatibility to IO.
+ */
 static VALUE
 strio_0(self)
     VALUE self;
@@ -282,6 +312,9 @@ strio_0(self)
     return INT2FIX(0);
 }
 
+/*
+ * Returns the argument unchanged.  Just for compatibility to IO.
+ */
 static VALUE
 strio_first(self, arg)
     VALUE self, arg;
@@ -290,6 +323,9 @@ strio_first(self, arg)
     return arg;
 }
 
+/*
+ * Raises NotImplementedError.
+ */
 static VALUE
 strio_unimpl(argc, argv, self)
     int argc;
@@ -301,6 +337,11 @@ strio_unimpl(argc, argv, self)
     return Qnil;		/* not reached */
 }
 
+/*
+ * call-seq: strio.string     -> string
+ *
+ * Returns underlying String object, the subject of IO.
+ */
 static VALUE
 strio_get_string(self)
     VALUE self;
@@ -308,6 +349,12 @@ strio_get_string(self)
     return StringIO(self)->string;
 }
 
+/*
+ * call-seq:
+ *   strio.string = string  -> string
+ *
+ * Changes underlying String object, the subject of IO.
+ */
 static VALUE
 strio_set_string(self, string)
     VALUE self, string;
@@ -325,6 +372,13 @@ strio_set_string(self, string)
     return ptr->string = string;
 }
 
+/*
+ * call-seq:
+ *   strio.close  -> nil
+ *
+ * Closes strio.  The *strio* is unavailable for any further data 
+ * operations; an +IOError+ is raised if such an attempt is made.
+ */
 static VALUE
 strio_close(self)
     VALUE self;
@@ -335,9 +389,16 @@ strio_close(self)
     }
     ptr->string = Qnil;
     ptr->flags &= ~FMODE_READWRITE;
-    return self;
+    return Qnil;
 }
 
+/*
+ * call-seq:
+ *   strio.close_read    -> nil
+ *
+ * Closes the read end of a StringIO.  Will raise an +IOError+ if the
+ * *strio* is not readable.
+ */
 static VALUE
 strio_close_read(self)
     VALUE self;
@@ -349,9 +410,16 @@ strio_close_read(self)
     if (!((ptr->flags &= ~FMODE_READABLE) & FMODE_READWRITE)) {
 	ptr->string = Qnil;
     }
-    return self;
+    return Qnil;
 }
 
+/*
+ * call-seq:
+ *   strio.close_write    -> nil
+ *
+ * Closes the write end of a StringIO.  Will raise an  +IOError+ if the
+ * *strio* is not writeable.
+ */
 static VALUE
 strio_close_write(self)
     VALUE self;
@@ -363,9 +431,15 @@ strio_close_write(self)
     if (!((ptr->flags &= ~FMODE_WRITABLE) & FMODE_READWRITE)) {
 	ptr->string = Qnil;
     }
-    return self;
+    return Qnil;
 }
 
+/*
+ * call-seq:
+ *   strio.closed?    -> true or false
+ *
+ * Returns +true+ if *strio* is completely closed, +false+ otherwise.
+ */
 static VALUE
 strio_closed(self)
     VALUE self;
@@ -375,6 +449,12 @@ strio_closed(self)
     return Qtrue;
 }
 
+/*
+ * call-seq:
+ *   strio.closed_read?    -> true or false
+ *
+ * Returns +true+ if *strio* is not readable, +false+ otherwise.
+ */
 static VALUE
 strio_closed_read(self)
     VALUE self;
@@ -384,6 +464,12 @@ strio_closed_read(self)
     return Qtrue;
 }
 
+/*
+ * call-seq:
+ *   strio.closed_write?    -> true or false
+ *
+ * Returns +true+ if *strio* is not writable, +false+ otherwise.
+ */
 static VALUE
 strio_closed_write(self)
     VALUE self;
@@ -393,6 +479,14 @@ strio_closed_write(self)
     return Qtrue;
 }
 
+/*
+ * call-seq:
+ *   strio.eof     -> true or false
+ *   strio.eof?    -> true or false
+ *
+ * Returns true if *strio* is at end of file. The stringio must be  
+ * opened for reading or an +IOError+ will be raised.
+ */
 static VALUE
 strio_eof(self)
     VALUE self;
@@ -402,6 +496,7 @@ strio_eof(self)
     return Qtrue;
 }
 
+/* :nodoc: */
 static VALUE
 strio_copy(copy, orig)
     VALUE copy, orig;
@@ -420,6 +515,16 @@ strio_copy(copy, orig)
     return copy;
 }
 
+/*
+ * call-seq:
+ *   strio.lineno    -> integer
+ *
+ * Returns the current line number in *strio*. The stringio must be
+ * opened for reading. +lineno+ counts the number of times  +gets+ is
+ * called, rather than the number of newlines  encountered. The two
+ * values will differ if +gets+ is  called with a separator other than
+ * newline.  See also the  <code>$.</code> variable.
+ */
 static VALUE
 strio_get_lineno(self)
     VALUE self;
@@ -427,6 +532,13 @@ strio_get_lineno(self)
     return LONG2NUM(StringIO(self)->lineno);
 }
 
+/*
+ * call-seq:
+ *   strio.lineno = integer    -> integer
+ *
+ * Manually sets the current line number to the given value.
+ * <code>$.</code> is updated only on the next read.
+ */
 static VALUE
 strio_set_lineno(self, lineno)
     VALUE self, lineno;
@@ -435,14 +547,26 @@ strio_set_lineno(self, lineno)
     return lineno;
 }
 
+/* call-seq: strio.binmode -> true */
 #define strio_binmode strio_self
 
+/* call-seq: strio.fcntl */
 #define strio_fcntl strio_unimpl
 
+/* call-seq: strio.flush -> strio */
 #define strio_flush strio_self
 
+/* call-seq: strio.fsync -> 0 */
 #define strio_fsync strio_0
 
+/*
+ * call-seq:
+ *   strio.reopen(other_StrIO)     -> strio
+ *   strio.reopen(string, mode)    -> strio
+ *
+ * Reinitializes *strio* with the given <i>other_StrIO</i> or _string_ 
+ * and _mode_ (see StringIO#new).
+ */
 static VALUE
 strio_reopen(argc, argv, self)
     int argc;
@@ -456,6 +580,13 @@ strio_reopen(argc, argv, self)
     return strio_initialize(argc, argv, self);
 }
 
+/*
+ * call-seq:
+ *   strio.pos     -> integer
+ *   strio.tell    -> integer
+ *
+ * Returns the current offset (in bytes) of *strio*.
+ */
 static VALUE
 strio_get_pos(self)
     VALUE self;
@@ -463,6 +594,12 @@ strio_get_pos(self)
     return LONG2NUM(StringIO(self)->pos);
 }
 
+/*
+ * call-seq:
+ *   strio.pos = integer    -> integer
+ *
+ * Seeks to the given position (in bytes) in *strio*.
+ */
 static VALUE
 strio_set_pos(self, pos)
     VALUE self;
@@ -474,9 +611,17 @@ strio_set_pos(self, pos)
 	error_inval(0);
     }
     ptr->pos = p;
+    ptr->flags &= ~STRIO_EOF;
     return pos;
 }
 
+/*
+ * call-seq:
+ *   strio.rewind    -> 0
+ *
+ * Positions *strio* to the beginning of input, resetting
+ * +lineno+ to zero.
+ */
 static VALUE
 strio_rewind(self)
     VALUE self;
@@ -488,6 +633,13 @@ strio_rewind(self)
     return INT2FIX(0);
 }
 
+/*
+ * call-seq:
+ *   strio.seek(amount, whence=SEEK_SET) -> 0
+ *
+ * Seeks to a given offset _amount_ in the stream according to
+ * the value of _whence_ (see IO#seek).
+ */
 static VALUE
 strio_seek(argc, argv, self)
     int argc;
@@ -520,6 +672,12 @@ strio_seek(argc, argv, self)
     return INT2FIX(0);
 }
 
+/*
+ * call-seq:
+ *   strio.sync    -> true
+ *
+ * Returns +true+ always.
+ */
 static VALUE
 strio_get_sync(self)
     VALUE self;
@@ -528,10 +686,17 @@ strio_get_sync(self)
     return Qtrue;
 }
 
+/* call-seq: strio.sync = boolean -> boolean */
 #define strio_set_sync strio_first
 
 #define strio_tell strio_get_pos
 
+/*
+ * call-seq:
+ *   strio.each_byte {|byte| block }  -> strio
+ *
+ * See IO#each_byte.
+ */
 static VALUE
 strio_each_byte(self)
     VALUE self;
@@ -544,6 +709,12 @@ strio_each_byte(self)
     return Qnil;
 }
 
+/*
+ * call-seq:
+ *   strio.getc   -> fixnum or nil
+ *
+ * See IO#getc.
+ */
 static VALUE
 strio_getc(self)
     VALUE self;
@@ -558,6 +729,35 @@ strio_getc(self)
     return CHR2FIX(c);
 }
 
+static void
+strio_extend(ptr, pos, len)
+    struct StringIO *ptr;
+    long pos, len;
+{
+    long olen;
+
+    check_modifiable(ptr);
+    olen = RSTRING(ptr->string)->len;
+    if (pos + len > olen) {
+	rb_str_resize(ptr->string, pos + len);
+	if (pos > olen)
+	    MEMZERO(RSTRING(ptr->string)->ptr + olen, char, pos - olen);
+    }
+    else {
+	rb_str_modify(ptr->string);
+    }
+}
+
+/*
+ * call-seq:
+ *   strio.ungetc(integer)   -> nil
+ *
+ * Pushes back one character (passed as a parameter) onto *strio*
+ * such that a subsequent buffered read will return it.  Pushing back 
+ * behind the beginning of the buffer string is not possible.  Nothing
+ * will be done if such an attempt is made.
+ * In other case, there is no limitation for multiple pushbacks.
+ */
 static VALUE
 strio_ungetc(self, ch)
     VALUE self, ch;
@@ -567,18 +767,11 @@ strio_ungetc(self, ch)
     long len, pos = ptr->pos;
 
     if (cc != EOF && pos > 0) {
-	if ((len = RSTRING(ptr->string)->len) < pos ||
-	    (unsigned char)RSTRING(ptr->string)->ptr[pos - 1] !=
+	if ((len = RSTRING(ptr->string)->len) < pos-- ||
+	    (unsigned char)RSTRING(ptr->string)->ptr[pos] !=
 	    (unsigned char)cc) {
-	    check_modifiable(ptr);
-	    if (len < pos) {
-		rb_str_resize(ptr->string, pos);
-		MEMZERO(RSTRING(ptr->string)->ptr + len, char, pos - len - 1);
-	    }
-	    else {
-		rb_str_modify(ptr->string);
-	    }
-	    RSTRING(ptr->string)->ptr[pos - 1] = cc;
+	    strio_extend(ptr, pos, 1);
+	    RSTRING(ptr->string)->ptr[pos] = cc;
 	    OBJ_INFECT(ptr->string, self);
 	}
 	--ptr->pos;
@@ -587,6 +780,12 @@ strio_ungetc(self, ch)
     return Qnil;
 }
 
+/*
+ * call-seq:
+ *   strio.readchar   -> fixnum
+ *
+ * See IO#readchar.
+ */
 static VALUE
 strio_readchar(self)
     VALUE self;
@@ -673,8 +872,7 @@ strio_getline(argc, argv, ptr)
 	    }
 	}
 	s = p;
-	while (p = memchr(p, '\n', e - p)) {
-	    if (p == e) break;
+	while ((p = memchr(p, '\n', e - p)) && (p != e)) {
 	    if (*++p == '\n') {
 		e = p;
 		break;
@@ -683,7 +881,7 @@ strio_getline(argc, argv, ptr)
 	str = rb_str_substr(ptr->string, s - RSTRING(ptr->string)->ptr, e - s); 
     }
     else if (n == 1) {
-	if (p = memchr(s, RSTRING(str)->ptr[0], e - s)) {
+	if ((p = memchr(s, RSTRING(str)->ptr[0], e - s)) != 0) {
 	    e = p + 1;
 	}
 	str = rb_str_substr(ptr->string, ptr->pos, e - s);
@@ -714,6 +912,12 @@ strio_getline(argc, argv, ptr)
     return str;
 }
 
+/*
+ * call-seq:
+ *   strio.gets(sep_string=$/)   -> string or nil
+ *
+ * See IO#gets.
+ */
 static VALUE
 strio_gets(argc, argv, self)
     int argc;
@@ -726,6 +930,12 @@ strio_gets(argc, argv, self)
     return str;
 }
 
+/*
+ * call-seq:
+ *   strio.readline(sep_string=$/)   -> string
+ *
+ * See IO#readline.
+ */
 static VALUE
 strio_readline(argc, argv, self)
     int argc;
@@ -737,6 +947,13 @@ strio_readline(argc, argv, self)
     return line;
 }
 
+/*
+ * call-seq:
+ *   strio.each(sep_string=$/)      {|line| block }  -> strio
+ *   strio.each_line(sep_string=$/) {|line| block }  -> strio
+ *
+ * See IO#each.
+ */
 static VALUE
 strio_each(argc, argv, self)
     int argc;
@@ -752,6 +969,12 @@ strio_each(argc, argv, self)
     return self;
 }
 
+/*
+ * call-seq:
+ *   strio.readlines(sep_string=$/)  ->   array
+ *
+ * See IO#readlines.
+ */
 static VALUE
 strio_readlines(argc, argv, self)
     int argc;
@@ -766,31 +989,37 @@ strio_readlines(argc, argv, self)
     return ary;
 }
 
+/*
+ * call-seq:
+ *   strio.write(string)    -> integer
+ *   strio.syswrite(string) -> integer
+ *
+ * Appends the given string to the underlying buffer string of *strio*.
+ * The stream must be opened for writing.  If the argument is not a
+ * string, it will be converted to a string using <code>to_s</code>.
+ * Returns the number of bytes written.  See IO#write.
+ */
 static VALUE
 strio_write(self, str)
     VALUE self, str;
 {
     struct StringIO *ptr = writable(StringIO(self));
-    long len;
+    long len, olen;
 
     if (TYPE(str) != T_STRING)
 	str = rb_obj_as_string(str);
     len = RSTRING(str)->len;
     if (!len) return INT2FIX(0);
     check_modifiable(ptr);
-    if (ptr->flags & STRIO_APPEND) {
-	ptr->pos = RSTRING(ptr->string)->len;
+    olen = RSTRING(ptr->string)->len;
+    if (ptr->flags & FMODE_APPEND) {
+	ptr->pos = olen;
     }
-    if (ptr->pos == RSTRING(ptr->string)->len) {
+    if (ptr->pos == olen) {
 	rb_str_cat(ptr->string, RSTRING(str)->ptr, len);
     }
     else {
-	if (ptr->pos + len > RSTRING(ptr->string)->len) {
-	    rb_str_resize(ptr->string, ptr->pos + len);
-	}
-	else {
-	    rb_str_modify(ptr->string);
-	}
+	strio_extend(ptr, ptr->pos, len);
 	rb_str_update(ptr->string, ptr->pos, len, str);
     }
     OBJ_INFECT(ptr->string, self);
@@ -798,36 +1027,70 @@ strio_write(self, str)
     return LONG2NUM(len);
 }
 
+/*
+ * call-seq:
+ *   strio << obj     -> strio
+ *
+ * See IO#<<.
+ */
 #define strio_addstr rb_io_addstr
 
+/*
+ * call-seq:
+ *   strio.print()             -> nil
+ *   strio.print(obj, ...)     -> nil
+ *
+ * See IO#print.
+ */
 #define strio_print rb_io_print
 
+/*
+ * call-seq:
+ *   strio.printf(format_string [, obj, ...] )   -> nil
+ *
+ * See IO#printf.
+ */
 #define strio_printf rb_io_printf
 
+/*
+ * call-seq:
+ *   strio.putc(obj)    -> obj
+ *
+ * See IO#putc.
+ */
 static VALUE
 strio_putc(self, ch)
     VALUE self, ch;
 {
     struct StringIO *ptr = writable(StringIO(self));
     int c = NUM2CHR(ch);
+    long olen;
 
     check_modifiable(ptr);
-    if (ptr->flags & STRIO_APPEND) {
-	ptr->pos = RSTRING(ptr->string)->len;
-    }
-    if (ptr->pos >= RSTRING(ptr->string)->len) {
-	rb_str_resize(ptr->string, ptr->pos + 1);
-    }
-    else {
-	rb_str_modify(ptr->string);
+    olen = RSTRING(ptr->string)->len;
+    if (ptr->flags & FMODE_APPEND) {
+	ptr->pos = olen;
     }
+    strio_extend(ptr, ptr->pos, 1);
     RSTRING(ptr->string)->ptr[ptr->pos++] = c;
     OBJ_INFECT(ptr->string, self);
     return ch;
 }
 
+/*
+ * call-seq:
+ *   strio.puts(obj, ...)    -> nil
+ *
+ * See IO#puts.
+ */
 #define strio_puts rb_io_puts
 
+/*
+ * call-seq:
+ *   strio.read([length [, buffer]])    -> string, buffer, or nil
+ *
+ * See IO#read.
+ */
 static VALUE
 strio_read(argc, argv, self)
     int argc;
@@ -888,7 +1151,7 @@ strio_read(argc, argv, self)
 	long rest = RSTRING(ptr->string)->len - ptr->pos;
 	if (len > rest) len = rest;
 	rb_str_resize(str, len);
-	MEMCPY(RSTRING(str)->ptr, RSTRING(ptr->string)->ptr, char, len);
+	MEMCPY(RSTRING(str)->ptr, RSTRING(ptr->string)->ptr + ptr->pos, char, len);
     }
     if (NIL_P(str)) {
 	if (!(ptr->flags & STRIO_EOF)) str = rb_str_new(0, 0);
@@ -901,6 +1164,13 @@ strio_read(argc, argv, self)
     return str;
 }
 
+/*
+ * call-seq:
+ *   strio.sysread(integer[, outbuf])    -> string
+ *
+ * Similar to #read, but raises +EOFError+ at end of string instead of
+ * returning +nil+, as well as IO#sysread does.
+ */
 static VALUE
 strio_sysread(argc, argv, self)
     int argc;
@@ -916,14 +1186,29 @@ strio_sysread(argc, argv, self)
 
 #define strio_syswrite strio_write
 
+/* call-seq: strio.path -> nil */
 #define strio_path strio_nil
 
+/*
+ * call-seq:
+ *   strio.isatty -> nil
+ *   strio.tty? -> nil
+ *
+ */
 #define strio_isatty strio_false
 
+/* call-seq: strio.pid -> nil */
 #define strio_pid strio_nil
 
+/* call-seq: strio.fileno -> nil */
 #define strio_fileno strio_nil
 
+/*
+ * call-seq:
+ *   strio.size   -> integer
+ *
+ * Returns the size of the buffer string.
+ */
 static VALUE
 strio_size(self)
     VALUE self;
@@ -935,19 +1220,33 @@ strio_size(self)
     return ULONG2NUM(RSTRING(string)->len);
 }
 
+/*
+ * call-seq:
+ *   strio.truncate(integer)    -> 0
+ *
+ * Truncates the buffer string to at most _integer_ bytes. The *strio*
+ * must be opened for writing.
+ */
 static VALUE
 strio_truncate(self, len)
     VALUE self, len;
 {
     VALUE string = writable(StringIO(self))->string;
     long l = NUM2LONG(len);
+    long plen = RSTRING(string)->len;
     if (l < 0) {
 	error_inval("negative legnth");
     }
     rb_str_resize(string, l);
+    if (plen < l) {
+	MEMZERO(RSTRING(string)->ptr + plen, char, l - plen);
+    }
     return len;
 }
 
+/*
+ * Pseudo I/O on String object.
+ */
 void
 Init_stringio()
 {
