@@ -9,6 +9,9 @@ module TkConsole
 
   TkCommandNames = ['console'.freeze].freeze
 
+  def self.create
+    TkCore::INTERP._create_console
+  end
   def self.title(str=None)
     tk_call 'console', str
   end
@@ -24,6 +27,6 @@ module TkConsole
     # I have no idea to support a Ruby script seamlessly.
     #
     _fromUTF8(tk_call_without_enc('console', 'eval', 
-				  _get_eval_enc_str(tcl_script)))
+                                  _get_eval_enc_str(tcl_script)))
   end
 end
