@@ -20,11 +20,11 @@ module Tk
   module Tcllib
     class IP_Entry < TkEntry
       def self.package_version
-	begin
-	  TkPackage.require('ipentry')
-	rescue
-	  ''
-	end
+        begin
+          TkPackage.require('ipentry')
+        rescue
+          ''
+        end
       end
     end
     IPEntry = IP_Entry
@@ -38,9 +38,10 @@ class Tk::Tcllib::IP_Entry
 
   def create_self(keys)
     if keys and keys != None
-      tk_call_without_enc('::ipentry::ipentry', @path, *hash_kv(keys, true))
+      tk_call_without_enc(self.class::TkCommandNames[0], @path, 
+                          *hash_kv(keys, true))
     else
-      tk_call_without_enc('::ipentry::ipentry', @path)
+      tk_call_without_enc(self.class::TkCommandNames[0], @path)
     end
   end
   private :create_self
