@@ -3,7 +3,7 @@
  *
  *   Copyright (C) UENO Katsuhiro 2000-2003
  *
- * $Id: zlib.c,v 1.7.2.4 2004/06/23 09:28:28 matz Exp $
+ * $Id: zlib.c,v 1.7.2.16 2005/04/18 02:34:20 ocean Exp $
  */
 
 #include <ruby.h>
@@ -55,7 +55,7 @@ static void zstream_reset_input _((struct zstream*));
 static void zstream_passthrough_input _((struct zstream*));
 static VALUE zstream_detach_input _((struct zstream*));
 static void zstream_reset _((struct zstream*));
-static void zstream_end _((struct zstream*));
+static VALUE zstream_end _((struct zstream*));
 static void zstream_run _((struct zstream*, Bytef*, uInt, int));
 static VALUE zstream_sync _((struct zstream*, Bytef*, uInt));
 static void zstream_mark _((struct zstream*));
@@ -80,7 +80,8 @@ static VALUE rb_zstream_closed_p _((VALUE));
 
 static VALUE rb_deflate_s_allocate _((VALUE));
 static VALUE rb_deflate_initialize _((int, VALUE*, VALUE));
-static VALUE rb_deflate_clone _((VALUE));
+static VALUE rb_deflate_init_copy _((VALUE, VALUE));
+static VALUE deflate_run _((VALUE));
 static VALUE rb_deflate_s_deflate _((int, VALUE*, VALUE));
 static void do_deflate _((struct zstream*, VALUE, int));
 static VALUE rb_deflate_deflate _((int, VALUE*, VALUE));
@@ -89,6 +90,7 @@ static VALUE rb_deflate_flush _((int, VALUE*, VALUE));
 static VALUE rb_deflate_params _((VALUE, VALUE, VALUE));
 static VALUE rb_deflate_set_dictionary _((VALUE, VALUE));
 
+static VALUE inflate_run _((VALUE));
 static VALUE rb_inflate_s_allocate _((VALUE));
 static VALUE rb_inflate_initialize _((int, VALUE*, VALUE));
 static VALUE rb_inflate_s_inflate _((VALUE, VALUE));
@@ -119,6 +121,7 @@ static void gzfile_read_header _((struct gzfile*));
 static void gzfile_check_footer _((struct gzfile*));
 static void gzfile_write _((struct gzfile*, Bytef*, uInt));
 static long gzfile_read_more _((struct gzfile*));
+static void gzfile_calc_crc _((struct gzfile*, VALUE));
 static VALUE gzfile_read _((struct gzfile*, int));
 static VALUE gzfile_read_all _((struct gzfile*));
 static void gzfile_ungetc _((struct gzfile*, int));
@@ -361,12 +364,14 @@ struct zstream {
 #define ZSTREAM_FLAG_IN_STREAM  0x2
 #define ZSTREAM_FLAG_FINISHED   0x4
 #define ZSTREAM_FLAG_FINALIZE   0x8
-#define ZSTREAM_FLAG_UNUSED     0x10
+#define ZSTREAM_FLAG_CLOSED     0x10
+#define ZSTREAM_FLAG_UNUSED     0x20
 
 #define ZSTREAM_READY(z)       ((z)->flags |= ZSTREAM_FLAG_READY)
 #define ZSTREAM_IS_READY(z)    ((z)->flags & ZSTREAM_FLAG_READY)
 #define ZSTREAM_IS_FINISHED(z) ((z)->flags & ZSTREAM_FLAG_FINISHED)
 #define ZSTREAM_IS_FINALIZE(z) ((z)->flags & ZSTREAM_FLAG_FINALIZE)
+#define ZSTREAM_IS_CLOSED(z)   ((z)->flags & ZSTREAM_FLAG_CLOSED)
 
 /* I think that more better value should be found,
    but I gave up finding it. B) */
@@ -434,6 +439,7 @@ zstream_expand_buffer(z)
 	z->buf_filled = 0;
 	z->stream.next_out = RSTRING(z->buf)->ptr;
 	z->stream.avail_out = ZSTREAM_INITIAL_BUFSIZE;
+	RBASIC(z->buf)->klass = 0;
 	return;
     }
 
@@ -465,6 +471,7 @@ zstream_expand_buffer_into(z, size)
 	z->buf_filled = 0;
 	z->stream.next_out = RSTRING(z->buf)->ptr;
 	z->stream.avail_out = size;
+	RBASIC(z->buf)->klass = 0;
     }
     else if (z->stream.avail_out != size) {
 	rb_str_resize(z->buf, z->buf_filled + size);
@@ -485,6 +492,7 @@ zstream_append_buffer(z, src, len)
 	z->buf_filled = len;
 	z->stream.next_out = RSTRING(z->buf)->ptr;
 	z->stream.avail_out = 0;
+	RBASIC(z->buf)->klass = 0;
 	return;
     }
 
@@ -520,6 +528,7 @@ zstream_detach_buffer(z)
     else {
 	dst = z->buf;
 	rb_str_resize(dst, z->buf_filled);
+	RBASIC(dst)->klass = rb_cString;
     }
 
     z->buf = Qnil;
@@ -541,6 +550,7 @@ zstream_shift_buffer(z, len)
     }
 
     dst = rb_str_substr(z->buf, 0, len);
+    RBASIC(dst)->klass = rb_cString;
     z->buf_filled -= len;
     memmove(RSTRING(z->buf)->ptr, RSTRING(z->buf)->ptr + len,
 	    z->buf_filled);
@@ -582,6 +592,7 @@ zstream_append_input(z, src, len)
     if (NIL_P(z->input)) {
 	z->input = rb_str_buf_new(len);
 	rb_str_buf_cat(z->input, src, len);
+	RBASIC(z->input)->klass = 0;
     }
     else {
 	rb_str_buf_cat(z->input, src, len);
@@ -629,7 +640,13 @@ zstream_detach_input(z)
 {
     VALUE dst;
 
-    dst = NIL_P(z->input) ? rb_str_new(0, 0) : z->input;
+    if (NIL_P(z->input)) {
+	dst = rb_str_new(0, 0);
+    }
+    else {
+	dst = z->input;
+	RBASIC(dst)->klass = rb_cString;
+    }
     z->input = Qnil;
     return dst;
 }
@@ -652,7 +669,7 @@ zstream_reset(z)
     zstream_reset_input(z);
 }
 
-static void
+static VALUE
 zstream_end(z)
     struct zstream *z;
 {
@@ -662,7 +679,7 @@ zstream_end(z)
 	if (RTEST(ruby_debug)) {
 	    rb_warning("attempt to close uninitialized zstream; ignored.");
 	}
-	return;
+	return Qnil;
     }
     if (z->flags & ZSTREAM_FLAG_IN_STREAM) {
 	if (RTEST(ruby_debug)) {
@@ -677,6 +694,7 @@ zstream_end(z)
 	raise_zlib_error(err, z->stream.msg);
     }
     z->flags = 0;
+    return Qnil;
 }
 
 static void
@@ -688,15 +706,20 @@ zstream_run(z, src, len, flush)
 {
     uInt n;
     int err;
+    volatile VALUE guard;
 
-    if (NIL_P(z->input)) {
-	z->stream.next_in = src;
-	z->stream.avail_in = len;
+    if (NIL_P(z->input) && len == 0) {
+	z->stream.next_in = "";
+	z->stream.avail_in = 0;
     }
     else {
 	zstream_append_input(z, src, len);
 	z->stream.next_in = RSTRING(z->input)->ptr;
 	z->stream.avail_in = RSTRING(z->input)->len;
+	/* keep reference to `z->input' so as not to be garbage collected
+	   after zstream_reset_input() and prevent `z->stream.next_in'
+	   from dangling. */
+	guard = z->input;
     }
 
     if (z->stream.avail_out == 0) {
@@ -1097,7 +1120,7 @@ static VALUE
 rb_deflate_s_allocate(klass)
     VALUE klass;
 {
-	return zstream_deflate_new(klass);
+    return zstream_deflate_new(klass);
 }
 
 /*
@@ -1137,26 +1160,31 @@ rb_deflate_initialize(argc, argv, obj)
  * Duplicates the deflate stream.
  */
 static VALUE
-rb_deflate_clone(obj)
-    VALUE obj;
+rb_deflate_init_copy(self, orig)
+    VALUE self, orig;
 {
-    struct zstream *z = get_zstream(obj);
-    struct zstream *z2;
-    VALUE clone;
+    struct zstream *z1 = get_zstream(self);
+    struct zstream *z2 = get_zstream(orig);
     int err;
 
-    clone = zstream_deflate_new(rb_class_of(obj));
-    Data_Get_Struct(clone, struct zstream, z2);
-
-    err = deflateCopy(&z2->stream, &z->stream);
+    err = deflateCopy(&z1->stream, &z2->stream);
     if (err != Z_OK) {
 	raise_zlib_error(err, 0);
     }
+    z1->flags = z2->flags;
 
-    z2->flags = z->flags;
-    CLONESETUP(clone, obj);
-    OBJ_INFECT(clone, obj);
-    return clone;
+    return self;
+}
+
+static VALUE
+deflate_run(args)
+    VALUE args;
+{
+    struct zstream *z = (struct zstream *)((VALUE *)args)[0];
+    VALUE src = ((VALUE *)args)[1];
+
+    zstream_run(z, RSTRING(src)->ptr, RSTRING(src)->len, Z_FINISH);
+    return zstream_detach_buffer(z);
 }
 
 /*
@@ -1186,22 +1214,23 @@ rb_deflate_s_deflate(argc, argv, klass)
     VALUE klass;
 {
     struct zstream z;
-    VALUE src, level, dst;
-    int err;
+    VALUE src, level, dst, args[2];
+    int err, lev;
 
     rb_scan_args(argc, argv, "11", &src, &level);
 
+    lev = ARG_LEVEL(level);
+    StringValue(src);
     zstream_init_deflate(&z);
-    err = deflateInit(&z.stream, ARG_LEVEL(level));
+    err = deflateInit(&z.stream, lev);
     if (err != Z_OK) {
 	raise_zlib_error(err, z.stream.msg);
     }
     ZSTREAM_READY(&z);
 
-    StringValue(src);
-    zstream_run(&z, RSTRING(src)->ptr, RSTRING(src)->len, Z_FINISH);
-    dst = zstream_detach_buffer(&z);
-    zstream_end(&z);
+    args[0] = (VALUE)&z;
+    args[1] = src;
+    dst = rb_ensure(deflate_run, (VALUE)args, zstream_end, (VALUE)&z);
 
     OBJ_INFECT(dst, src);
     return dst;
@@ -1414,6 +1443,18 @@ rb_inflate_initialize(argc, argv, obj)
     return obj;
 }
 
+static VALUE
+inflate_run(args)
+    VALUE args;
+{
+    struct zstream *z = (struct zstream *)((VALUE *)args)[0];
+    VALUE src = ((VALUE *)args)[1];
+
+    zstream_run(z, RSTRING(src)->ptr, RSTRING(src)->len, Z_SYNC_FLUSH);
+    zstream_run(z, "", 0, Z_FINISH);  /* for checking errors */
+    return zstream_detach_buffer(z);
+}
+
 /*
  * call-seq: Zlib::Inflate.inflate(string)
  *
@@ -1436,9 +1477,10 @@ rb_inflate_s_inflate(obj, src)
     VALUE obj, src;
 {
     struct zstream z;
-    VALUE dst;
+    VALUE dst, args[2];
     int err;
 
+    StringValue(src);
     zstream_init_inflate(&z);
     err = inflateInit(&z.stream);
     if (err != Z_OK) {
@@ -1446,11 +1488,9 @@ rb_inflate_s_inflate(obj, src)
     }
     ZSTREAM_READY(&z);
 
-    StringValue(src);
-    zstream_run(&z, RSTRING(src)->ptr, RSTRING(src)->len, Z_SYNC_FLUSH);
-    zstream_run(&z, "", 0, Z_FINISH);  /* for checking errors */
-    dst = zstream_detach_buffer(&z);
-    zstream_end(&z);
+    args[0] = (VALUE)&z;
+    args[1] = src;
+    dst = rb_ensure(inflate_run, (VALUE)args, zstream_end, (VALUE)&z);
 
     OBJ_INFECT(dst, src);
     return dst;
@@ -1757,6 +1797,7 @@ gzfile_close(gz, closeflag)
     int closeflag;
 {
     VALUE io = gz->io;
+
     gz->end(gz);
     gz->io = Qnil;
     gz->orig_name = Qnil;
@@ -2046,7 +2087,7 @@ static long
 gzfile_read_more(gz)
     struct gzfile *gz;
 {
-    VALUE str;
+    volatile VALUE str;
 
     while (!ZSTREAM_IS_FINISHED(&gz->z)) {
 	str = gzfile_read_raw(gz);
@@ -2065,6 +2106,21 @@ gzfile_read_more(gz)
     return gz->z.buf_filled;
 }
 
+static void
+gzfile_calc_crc(gz, str)
+    struct gzfile *gz;
+    VALUE str;
+{
+    if (RSTRING(str)->len <= gz->ungetc) {
+	gz->ungetc -= RSTRING(str)->len;
+    }
+    else {
+	gz->crc = crc32(gz->crc, RSTRING(str)->ptr + gz->ungetc,
+			RSTRING(str)->len - gz->ungetc);
+	gz->ungetc = 0;
+    }
+}
+
 static VALUE
 gzfile_read(gz, len)
     struct gzfile *gz;
@@ -2072,7 +2128,10 @@ gzfile_read(gz, len)
 {
     VALUE dst;
 
-    if (len <= 0) return Qnil;
+    if (len < 0)
+        rb_raise(rb_eArgError, "negative length %d given", len);
+    if (len == 0)
+	return rb_str_new(0, 0);
     while (!ZSTREAM_IS_FINISHED(&gz->z) && gz->z.buf_filled < len) {
 	gzfile_read_more(gz);
     }
@@ -2084,13 +2143,7 @@ gzfile_read(gz, len)
     }
 
     dst = zstream_shift_buffer(&gz->z, len);
-    if (RSTRING(dst)->len <= gz->ungetc) {
-	gz->ungetc -= RSTRING(dst)->len;
-    }
-    else {
-	gz->crc = crc32(gz->crc, RSTRING(dst)->ptr + gz->ungetc,
-			RSTRING(dst)->len - gz->ungetc);
-    }
+    gzfile_calc_crc(gz, dst);
 
     OBJ_TAINT(dst);  /* for safe */
     return dst;
@@ -2113,13 +2166,7 @@ gzfile_read_all(gz)
     }
 
     dst = zstream_detach_buffer(&gz->z);
-    if (RSTRING(dst)->len <= gz->ungetc) {
-	gz->ungetc -= RSTRING(dst)->len;
-    }
-    else {
-	gz->crc = crc32(gz->crc, RSTRING(dst)->ptr + gz->ungetc,
-			RSTRING(dst)->len - gz->ungetc);
-    }
+    gzfile_calc_crc(gz, dst);
 
     OBJ_TAINT(dst);  /* for safe */
     return dst;
@@ -2149,6 +2196,9 @@ gzfile_writer_end(gz)
 {
     int aborted;
 
+    if (ZSTREAM_IS_CLOSED(&gz->z)) return;
+    gz->z.flags |= ZSTREAM_FLAG_CLOSED;
+
     if (!(gz->z.flags & GZFILE_FLAG_HEADER_FINISHED)) {
 	gzfile_make_header(gz);
     }
@@ -2157,8 +2207,9 @@ gzfile_writer_end(gz)
     gzfile_make_footer(gz);
 
     if (ZSTREAM_IS_FINALIZE(&gz->z)) {
+	if (NIL_P(gz->io)) return;
 	rb_warn("Zlib::GzipWriter object must be closed explicitly.");
-	if (OBJ_IS_FREED(gz->io)) {
+	if (!SPECIAL_CONST_P(gz->io) && OBJ_IS_FREED(gz->io)) {
 	    aborted = 1;
 	}
 	else {
@@ -2178,6 +2229,9 @@ static void
 gzfile_reader_end(gz)
     struct gzfile *gz;
 {
+    if (ZSTREAM_IS_CLOSED(&gz->z)) return;
+    gz->z.flags |= ZSTREAM_FLAG_CLOSED;
+
     if (GZFILE_IS_FINISHED(gz)
 	&& !ZSTREAM_IS_FINALIZE(&gz->z)
 	&& !(gz->z.flags & GZFILE_FLAG_FOOTER_FINISHED)) {
@@ -2946,7 +3000,7 @@ rb_gzreader_readchar(obj)
     VALUE dst;
     dst = rb_gzreader_getc(obj);
     if (NIL_P(dst)) {
-	rb_raise(rb_eEOFError, "End of file reached");
+	rb_raise(rb_eEOFError, "end of file reached");
     }
     return dst;
 }
@@ -2995,8 +3049,7 @@ gzreader_skip_linebreaks(gz)
     while (n++, *(p++) == '\n') {
 	if (n >= gz->z.buf_filled) {
 	    str = zstream_detach_buffer(&gz->z);
-	    gz->crc = crc32(gz->crc, RSTRING(str)->ptr,
-			    RSTRING(str)->len);
+	    gzfile_calc_crc(gz, str);
 	    while (gz->z.buf_filled == 0) {
 		if (GZFILE_IS_FINISHED(gz)) return;
 		gzfile_read_more(gz);
@@ -3007,7 +3060,17 @@ gzreader_skip_linebreaks(gz)
     }
 
     str = zstream_shift_buffer(&gz->z, n - 1);
-    gz->crc = crc32(gz->crc, RSTRING(str)->ptr, RSTRING(str)->len);
+    gzfile_calc_crc(gz, str);
+}
+
+static void
+rscheck(rsptr, rslen, rs)
+    char *rsptr;
+    long rslen;
+    VALUE rs;
+{
+    if (RSTRING(rs)->ptr != rsptr && RSTRING(rs)->len != rslen)
+	rb_raise(rb_eRuntimeError, "rs modified");
 }
 
 static VALUE
@@ -3017,8 +3080,9 @@ gzreader_gets(argc, argv, obj)
     VALUE obj;
 {
     struct gzfile *gz = get_gzfile(obj);
-    VALUE rs, dst;
-    char *rsptr, *p;
+    volatile VALUE rs;
+    VALUE dst;
+    char *rsptr, *p, *res;
     long rslen, n;
     int rspara;
 
@@ -3060,16 +3124,24 @@ gzreader_gets(argc, argv, obj)
 	gzfile_read_more(gz);
     }
 
-    n = rslen;
     p = RSTRING(gz->z.buf)->ptr;
+    n = rslen;
     for (;;) {
 	if (n > gz->z.buf_filled) {
 	    if (ZSTREAM_IS_FINISHED(&gz->z)) break;
 	    gzfile_read_more(gz);
 	    p = RSTRING(gz->z.buf)->ptr + n - rslen;
 	}
-	if (memcmp(p, rsptr, rslen) == 0) break;
-	p++, n++;
+	if (!rspara) rscheck(rsptr, rslen, rs);
+	res = memchr(p, rsptr[0], (gz->z.buf_filled - n + 1));
+	if (!res) {
+	    n = gz->z.buf_filled + 1;
+	} else {
+	    n += (long)(res - p);
+	    p = res;
+	    if (rslen == 1 || memcmp(p, rsptr, rslen) == 0) break;
+	    p++, n++;
+	}
     }
 
     gz->lineno++;
@@ -3110,7 +3182,7 @@ rb_gzreader_readline(argc, argv, obj)
     VALUE dst;
     dst = rb_gzreader_gets(argc, argv, obj);
     if (NIL_P(dst)) {
-	rb_raise(rb_eEOFError, "End of file reached");
+	rb_raise(rb_eEOFError, "end of file reached");
     }
     return dst;
 }
@@ -3260,7 +3332,7 @@ void Init_zlib()
     cZStream = rb_define_class_under(mZlib, "ZStream", rb_cObject);
     rb_undef_alloc_func(cZStream);
     rb_define_method(cZStream, "avail_out", rb_zstream_avail_out, 0);
-    rb_define_method(cZStream, "avail_out=", rb_zstream_set_avail_out, 0);
+    rb_define_method(cZStream, "avail_out=", rb_zstream_set_avail_out, 1);
     rb_define_method(cZStream, "avail_in", rb_zstream_avail_in, 0);
     rb_define_method(cZStream, "total_in", rb_zstream_total_in, 0);
     rb_define_method(cZStream, "total_out", rb_zstream_total_out, 0);
@@ -3285,7 +3357,7 @@ void Init_zlib()
     rb_define_singleton_method(cDeflate, "deflate", rb_deflate_s_deflate, -1);
     rb_define_alloc_func(cDeflate, rb_deflate_s_allocate);
     rb_define_method(cDeflate, "initialize", rb_deflate_initialize, -1);
-    rb_define_method(cDeflate, "clone", rb_deflate_clone, 0);
+    rb_define_method(cDeflate, "initialize_copy", rb_deflate_init_copy, 0);
     rb_define_method(cDeflate, "deflate", rb_deflate_deflate, -1);
     rb_define_method(cDeflate, "<<", rb_deflate_addstr, 1);
     rb_define_method(cDeflate, "flush", rb_deflate_flush, -1);
