@@ -2,8 +2,8 @@
 
   file.c -
 
-  $Author: nobu $
-  $Date: 2004/04/14 04:06:51 $
+  $Author: eban $
+  $Date: 2005/04/11 13:26:32 $
   created at: Mon Nov 15 12:24:34 JST 1993
 
   Copyright (C) 1993-2003 Yukihiro Matsumoto
@@ -64,8 +64,11 @@ char *strrchr _((const char*,const char));
 #include <sys/mkdev.h>
 #endif
 
-#ifndef HAVE_LSTAT
-#define lstat(path,st) stat(path,st)
+#if !defined HAVE_LSTAT && !defined lstat
+#define lstat stat
+#endif
+#if !HAVE_FSEEKO && !defined(fseeko)
+# define fseeko  fseek
 #endif
 
 VALUE rb_cFile;
@@ -93,7 +96,7 @@ apply2files(func, vargs, arg)
 
 /*
  *  call-seq:
- *     file.path -> file_ame
+ *     file.path -> filename
  *  
  *  Returns the pathname used to create <i>file</i> as a string. Does
  *  not normalize the name.
@@ -110,6 +113,7 @@ rb_file_path(obj)
     OpenFile *fptr;
 
     fptr = RFILE(rb_io_taint_check(obj))->fptr;
+    rb_io_check_initialized(fptr);
     if (!fptr->path) return Qnil;
     return rb_tainted_str_new2(fptr->path);
 }
@@ -192,6 +196,52 @@ rb_stat_dev(self)
     return INT2NUM(get_stat(self)->st_dev);
 }
 
+/*
+ *  call-seq:
+ *     stat.dev_major   => fixnum
+ *  
+ *  Returns the major part of <code>File_Stat#dev</code> or
+ *  <code>nil</code>.
+ *     
+ *     File.stat("/dev/fd1").dev_major   #=> 2
+ *     File.stat("/dev/tty").dev_major   #=> 5
+ */
+
+static VALUE
+rb_stat_dev_major(self)
+    VALUE self;
+{
+#if defined(major)
+    long dev = get_stat(self)->st_dev;
+    return ULONG2NUM(major(dev));
+#else
+    return Qnil;
+#endif
+}
+
+/*
+ *  call-seq:
+ *     stat.dev_minor   => fixnum
+ *  
+ *  Returns the minor part of <code>File_Stat#dev</code> or
+ *  <code>nil</code>.
+ *     
+ *     File.stat("/dev/fd1").dev_minor   #=> 1
+ *     File.stat("/dev/tty").dev_minor   #=> 0
+ */
+
+static VALUE
+rb_stat_dev_minor(self)
+    VALUE self;
+{
+#if defined(minor)
+    long dev = get_stat(self)->st_dev;
+    return ULONG2NUM(minor(dev));
+#else
+    return Qnil;
+#endif
+}
+
 
 /*
  *  call-seq:
@@ -1598,8 +1648,7 @@ rb_file_s_lchmod(argc, argv)
 {
     VALUE vmode;
     VALUE rest;
-    int mode;
-    long n;
+    long mode, n;
 
     rb_secure(2);
     rb_scan_args(argc, argv, "1*", &vmode, &rest);
@@ -1695,15 +1744,18 @@ rb_file_chown(obj, owner, group)
     VALUE obj, owner, group;
 {
     OpenFile *fptr;
+    int o, g;
 
     rb_secure(2);
+    o = NUM2INT(owner);
+    g = NUM2INT(group);
     GetOpenFile(obj, fptr);
 #if defined(DJGPP) || defined(__CYGWIN32__) || defined(_WIN32) || defined(__EMX__)
     if (!fptr->path) return Qnil;
-    if (chown(fptr->path, NUM2INT(owner), NUM2INT(group)) == -1)
+    if (chown(fptr->path, o, g) == -1)
 	rb_sys_fail(fptr->path);
 #else
-    if (fchown(fileno(fptr->f), NUM2INT(owner), NUM2INT(group)) == -1)
+    if (fchown(fileno(fptr->f), o, g) == -1)
 	rb_sys_fail(fptr->path);
 #endif
 
@@ -2514,11 +2566,11 @@ rb_file_s_basename(argc, argv)
     VALUE *argv;
 {
     VALUE fname, fext, basename;
-    char *name, *p, *ext = NULL;
+    char *name, *p;
     int f;
 
     if (rb_scan_args(argc, argv, "11", &fname, &fext) == 2) {
-	ext = StringValueCStr(fext);
+	StringValue(fext);
     }
     StringValue(fname);
     if (RSTRING(fname)->len == 0 || !*(name = RSTRING(fname)->ptr))
@@ -2534,7 +2586,7 @@ rb_file_s_basename(argc, argv)
 #endif
     }
     else if (!(p = strrdirsep(name))) {
-	if (NIL_P(fext) || !(f = rmext(name, ext))) {
+	if (NIL_P(fext) || !(f = rmext(name, StringValueCStr(fext)))) {
 	    f = chompdirsep(name) - name;
 	    if (f == RSTRING(fname)->len) return fname;
 	}
@@ -2542,7 +2594,7 @@ rb_file_s_basename(argc, argv)
     }
     else {
 	while (isdirsep(*p)) p++; /* skip last / */
-	if (NIL_P(fext) || !(f = rmext(p, ext))) {
+	if (NIL_P(fext) || !(f = rmext(p, StringValueCStr(fext)))) {
 	    f = chompdirsep(p) - p;
 	}
     }
@@ -2755,11 +2807,14 @@ static VALUE
 rb_file_s_truncate(klass, path, len)
     VALUE klass, path, len;
 {
+    off_t pos;
+
     rb_secure(2);
+    pos = NUM2OFFT(len);
     SafeStringValue(path);
 
 #ifdef HAVE_TRUNCATE
-    if (truncate(StringValueCStr(path), NUM2OFFT(len)) < 0)
+    if (truncate(StringValueCStr(path), pos) < 0)
 	rb_sys_fail(RSTRING(path)->ptr);
 #else
 # ifdef HAVE_CHSIZE
@@ -2775,7 +2830,7 @@ rb_file_s_truncate(klass, path, len)
 	    rb_sys_fail(RSTRING(path)->ptr);
 	}
 #  endif
-	if (chsize(tmpfd, NUM2OFFT(len)) < 0) {
+	if (chsize(tmpfd, pos) < 0) {
 	    close(tmpfd);
 	    rb_sys_fail(RSTRING(path)->ptr);
 	}
@@ -2807,18 +2862,24 @@ rb_file_truncate(obj, len)
     VALUE obj, len;
 {
     OpenFile *fptr;
+    FILE *f;
+    off_t pos;
 
     rb_secure(2);
+    pos = NUM2OFFT(len);
     GetOpenFile(obj, fptr);
     if (!(fptr->mode & FMODE_WRITABLE)) {
 	rb_raise(rb_eIOError, "not opened for writing");
     }
+    f = GetWriteFile(fptr);
+    fflush(f);
+    fseeko(f, (off_t)0, SEEK_CUR);
 #ifdef HAVE_TRUNCATE
-    if (ftruncate(fileno(fptr->f), NUM2OFFT(len)) < 0)
+    if (ftruncate(fileno(f), pos) < 0)
 	rb_sys_fail(fptr->path);
 #else
 # ifdef HAVE_CHSIZE
-    if (chsize(fileno(fptr->f), NUM2OFFT(len)) < 0)
+    if (chsize(fileno(f), pos) < 0)
 	rb_sys_fail(fptr->path);
 # else
     rb_notimplement();
@@ -2840,7 +2901,7 @@ rb_file_truncate(obj, len)
 #  define LOCK_UN 8
 # endif
 
-#if 0
+#if 1
 static int
 rb_thread_flock(fd, op, fptr)
     int fd, op;
@@ -2905,15 +2966,17 @@ rb_file_flock(obj, operation)
 {
 #ifndef __CHECKER__
     OpenFile *fptr;
+    int op;
 
     rb_secure(2);
+    op = NUM2INT(operation);
     GetOpenFile(obj, fptr);
 
     if (fptr->mode & FMODE_WRITABLE) {
 	fflush(GetWriteFile(fptr));
     }
   retry:
-    if (flock(fileno(fptr->f), NUM2INT(operation)) < 0) {
+    if (flock(fileno(fptr->f), op) < 0) {
         switch (errno) {
           case EAGAIN:
           case EACCES:
@@ -3884,6 +3947,7 @@ rb_find_file_ext(filep, ext)
 	if (rb_safe_level() >= 2 && OBJ_TAINTED(fname)) {
 	    rb_raise(rb_eSecurityError, "loading from unsafe file %s", f);
 	}
+	OBJ_FREEZE(fname);
 	f = StringValueCStr(fname);
 	*filep = fname;
     }
@@ -3892,6 +3956,7 @@ rb_find_file_ext(filep, ext)
 	for (i=0; ext[i]; i++) {
 	    fname = rb_str_dup(*filep);
 	    rb_str_cat2(fname, ext[i]);
+	    OBJ_FREEZE(fname);
 	    if (file_load_ok(StringValueCStr(fname))) {
 		*filep = fname;
 		return i+1;
@@ -3912,6 +3977,7 @@ rb_find_file_ext(filep, ext)
 	for (j=0; ext[j]; j++) {
 	    fname = rb_str_dup(*filep);
 	    rb_str_cat2(fname, ext[j]);
+	    OBJ_FREEZE(fname);
 	    found = dln_find_file(StringValueCStr(fname), path);
 	    if (found && file_load_ok(found)) {
 		*filep = fname;
@@ -3935,6 +4001,7 @@ rb_find_file(path)
 	if (rb_safe_level() >= 1 && OBJ_TAINTED(path)) {
 	    rb_raise(rb_eSecurityError, "loading from unsafe path %s", f);
 	}
+	OBJ_FREEZE(path);
 	f = StringValueCStr(path);
     }
 
@@ -3993,7 +4060,9 @@ rb_find_file(path)
 	rb_raise(rb_eSecurityError, "loading from unsafe file %s", f);
     }
     if (file_load_ok(f)) {
-	return rb_str_new2(f);
+	tmp = rb_str_new2(f);
+	OBJ_FREEZE(tmp);
+	return tmp;
     }
     return 0;
 }
@@ -4148,6 +4217,8 @@ Init_File()
     rb_define_method(rb_cStat, "<=>", rb_stat_cmp, 1);
 
     rb_define_method(rb_cStat, "dev", rb_stat_dev, 0);
+    rb_define_method(rb_cStat, "dev_major", rb_stat_dev_major, 0);
+    rb_define_method(rb_cStat, "dev_minor", rb_stat_dev_minor, 0);
     rb_define_method(rb_cStat, "ino", rb_stat_ino, 0);
     rb_define_method(rb_cStat, "mode", rb_stat_mode, 0);
     rb_define_method(rb_cStat, "nlink", rb_stat_nlink, 0);
