@@ -2,8 +2,8 @@
 
   hash.c -
 
-  $Author: eban $
-  $Date: 2004/05/13 05:27:24 $
+  $Author: matz $
+  $Date: 2005/02/28 02:45:20 $
   created at: Mon Nov 22 18:51:18 JST 1993
 
   Copyright (C) 1993-2003 Yukihiro Matsumoto
@@ -114,39 +114,77 @@ static struct st_hash_type objhash = {
     rb_any_hash,
 };
 
-struct rb_hash_foreach_arg {
-    VALUE hash;
-    enum st_retval (*func)();
-    VALUE arg;
+struct foreach_safe_arg {
+    st_table *tbl;
+    int (*func)();
+    st_data_t arg;
 };
 
 static int
-rb_hash_foreach_iter(key, value, arg)
-    VALUE key, value;
-    struct rb_hash_foreach_arg *arg;
+foreach_safe_i(key, value, arg)
+    st_data_t key, value;
+    struct foreach_safe_arg *arg;
 {
     int status;
-    st_table *tbl = RHASH(arg->hash)->tbl;
-    struct st_table_entry **bins = tbl->bins;
 
     if (key == Qundef) return ST_CONTINUE;
     status = (*arg->func)(key, value, arg->arg);
-    if (RHASH(arg->hash)->tbl != tbl || RHASH(arg->hash)->tbl->bins != bins){
-	rb_raise(rb_eIndexError, "rehash occurred during iteration");
+    if (status == ST_CONTINUE) {
+	return ST_CHECK;
     }
     return status;
 }
 
-static VALUE
-rb_hash_foreach_call(arg)
-    struct rb_hash_foreach_arg *arg;
+void
+st_foreach_safe(table, func, a)
+    st_table *table;
+    int (*func)();
+    st_data_t a;
 {
-    st_foreach(RHASH(arg->hash)->tbl, rb_hash_foreach_iter, (st_data_t)arg);
-    return Qnil;
+    struct foreach_safe_arg arg;
+
+    arg.tbl = table;
+    arg.func = func;
+    arg.arg = a;
+    if (st_foreach(table, foreach_safe_i, (st_data_t)&arg)) {
+	rb_raise(rb_eRuntimeError, "hash modified during iteration");
+    }
+}
+
+struct hash_foreach_arg {
+    VALUE hash;
+    int (*func)();
+    VALUE arg;
+};
+
+static int
+hash_foreach_iter(key, value, arg)
+    VALUE key, value;
+    struct hash_foreach_arg *arg;
+{
+    int status;
+    st_table *tbl;
+
+    tbl = RHASH(arg->hash)->tbl;    
+    if (key == Qundef) return ST_CONTINUE;
+    status = (*arg->func)(key, value, arg->arg);
+    if (RHASH(arg->hash)->tbl != tbl) {
+	rb_raise(rb_eRuntimeError, "rehash occurred during iteration");
+    }
+    switch (status) {
+      case ST_DELETE:
+ 	st_delete_safe(tbl, (st_data_t*)&key, 0, Qundef);
+	FL_SET(arg->hash, HASH_DELETED);
+      case ST_CONTINUE:
+ 	break;
+      case ST_STOP:
+ 	return ST_STOP;
+    }
+    return ST_CHECK;
 }
 
 static VALUE
-rb_hash_foreach_ensure(hash)
+hash_foreach_ensure(hash)
     VALUE hash;
 {
     RHASH(hash)->iter_lev--;
@@ -160,19 +198,29 @@ rb_hash_foreach_ensure(hash)
     return 0;
 }
 
-static int
+static VALUE
+hash_foreach_call(arg)
+    struct hash_foreach_arg *arg;
+{
+    if (st_foreach(RHASH(arg->hash)->tbl, hash_foreach_iter, (st_data_t)arg)) {
+ 	rb_raise(rb_eRuntimeError, "hash modified during iteration");
+    }
+    return Qnil;
+}
+
+void
 rb_hash_foreach(hash, func, farg)
     VALUE hash;
-    enum st_retval (*func)();
+    int (*func)();
     VALUE farg;
 {
-    struct rb_hash_foreach_arg arg;
+    struct hash_foreach_arg arg;
 
     RHASH(hash)->iter_lev++;
     arg.hash = hash;
     arg.func = func;
     arg.arg  = farg;
-    return rb_ensure(rb_hash_foreach_call, (VALUE)&arg, rb_hash_foreach_ensure, hash);
+    rb_ensure(hash_foreach_call, (VALUE)&arg, hash_foreach_ensure, hash);
 }
 
 static VALUE hash_alloc _((VALUE));
@@ -341,7 +389,7 @@ rb_hash_rehash(hash)
 
     rb_hash_modify(hash);
     tbl = st_init_table_with_size(&objhash, RHASH(hash)->tbl->num_entries);
-    st_foreach(RHASH(hash)->tbl, rb_hash_rehash_i, (st_data_t)tbl);
+    rb_hash_foreach(hash, rb_hash_rehash_i, (st_data_t)tbl);
     st_free_table(RHASH(hash)->tbl);
     RHASH(hash)->tbl = tbl;
 
@@ -553,7 +601,7 @@ rb_hash_index(hash, value)
     args[0] = value;
     args[1] = Qnil;
 
-    st_foreach(RHASH(hash)->tbl, index_i, (st_data_t)args);
+    rb_hash_foreach(hash, index_i, (st_data_t)args);
 
     return args[1];
 }
@@ -665,7 +713,7 @@ rb_hash_shift(hash)
 
     rb_hash_modify(hash);
     var.stop = 0;
-    st_foreach(RHASH(hash)->tbl, shift_i, (st_data_t)&var);
+    rb_hash_foreach(hash, shift_i, (st_data_t)&var);
 
     if (var.stop) {
 	return rb_assoc_new(var.key, var.val);
@@ -678,7 +726,7 @@ rb_hash_shift(hash)
     }
 }
 
-static enum st_retval
+static int
 delete_if_i(key, value, hash)
     VALUE key, value, hash;
 {
@@ -745,7 +793,7 @@ rb_hash_reject(hash)
     return rb_hash_delete_if(rb_obj_dup(hash));
 }
 
-static enum st_retval
+static int
 select_i(key, value, result)
     VALUE key, value, result;
 {
@@ -832,8 +880,12 @@ static VALUE
 rb_hash_clear(hash)
     VALUE hash;
 {
+    void *tmp;
+
     rb_hash_modify(hash);
-    st_foreach(RHASH(hash)->tbl, clear_i, 0);
+    if (RHASH(hash)->tbl->num_entries > 0) {
+	rb_hash_foreach(hash, clear_i, 0);
+    }
 
     return hash;
 }
@@ -900,7 +952,7 @@ rb_hash_replace(hash, hash2)
     hash2 = to_hash(hash2);
     if (hash == hash2) return hash;
     rb_hash_clear(hash);
-    st_foreach(RHASH(hash2)->tbl, replace_i, hash);
+    rb_hash_foreach(hash2, replace_i, hash);
     RHASH(hash)->ifnone = RHASH(hash2)->ifnone;
     if (FL_TEST(hash2, HASH_PROC_DEFAULT)) {
 	FL_SET(hash, HASH_PROC_DEFAULT);
@@ -952,7 +1004,7 @@ rb_hash_empty_p(hash)
     return Qfalse;
 }
 
-static enum st_retval
+static int
 each_value_i(key, value)
     VALUE key, value;
 {
@@ -985,7 +1037,7 @@ rb_hash_each_value(hash)
     return hash;
 }
 
-static enum st_retval
+static int
 each_key_i(key, value)
     VALUE key, value;
 {
@@ -1017,7 +1069,7 @@ rb_hash_each_key(hash)
     return hash;
 }
 
-static enum st_retval
+static int
 each_pair_i(key, value)
     VALUE key, value;
 {
@@ -1051,7 +1103,7 @@ rb_hash_each_pair(hash)
     return hash;
 }
 
-static enum st_retval
+static int
 each_i(key, value)
     VALUE key, value;
 {
@@ -1115,7 +1167,7 @@ rb_hash_to_a(hash)
     VALUE ary;
 
     ary = rb_ary_new();
-    st_foreach(RHASH(hash)->tbl, to_a_i, ary);
+    rb_hash_foreach(hash, to_a_i, ary);
     if (OBJ_TAINTED(hash)) OBJ_TAINT(ary);
 
     return ary;
@@ -1173,7 +1225,7 @@ inspect_hash(hash)
     VALUE str;
 
     str = rb_str_buf_new2("{");
-    st_foreach(RHASH(hash)->tbl, inspect_i, str);
+    rb_hash_foreach(hash, inspect_i, str);
     rb_str_buf_cat2(str, "}");
     OBJ_INFECT(str, hash);
 
@@ -1267,7 +1319,7 @@ rb_hash_keys(hash)
     VALUE ary;
 
     ary = rb_ary_new();
-    st_foreach(RHASH(hash)->tbl, keys_i, ary);
+    rb_hash_foreach(hash, keys_i, ary);
 
     return ary;
 }
@@ -1300,7 +1352,7 @@ rb_hash_values(hash)
     VALUE ary;
 
     ary = rb_ary_new();
-    st_foreach(RHASH(hash)->tbl, values_i, ary);
+    rb_hash_foreach(hash, values_i, ary);
 
     return ary;
 }
@@ -1365,7 +1417,7 @@ rb_hash_has_value(hash, val)
 
     data[0] = Qfalse;
     data[1] = val;
-    st_foreach(RHASH(hash)->tbl, rb_hash_search_value, (st_data_t)data);
+    rb_hash_foreach(hash, rb_hash_search_value, (st_data_t)data);
     return data[0];
 }
 
@@ -1417,7 +1469,7 @@ hash_equal(hash1, hash2, eql)
 
     data.tbl = RHASH(hash2)->tbl;
     data.result = Qtrue;
-    st_foreach(RHASH(hash1)->tbl, equal_i, (st_data_t)&data);
+    rb_hash_foreach(hash1, equal_i, (st_data_t)&data);
 
     return data.result;
 }
@@ -1448,22 +1500,6 @@ rb_hash_equal(hash1, hash2)
     return hash_equal(hash1, hash2, Qfalse);
 }
 
-/*
- *  call-seq:
- *     hsh.eql?(other_hash)    => true or false
- *  
- *  Returns true if two hashes are equal, i.e they have same key-value set,
- *  and same default values.
- *     
- */
-
-static VALUE
-rb_hash_eql(hash1, hash2)
-    VALUE hash1, hash2;
-{
-    return hash_equal(hash1, hash2, Qtrue);
-}
-
 static int
 rb_hash_invert_i(key, value, hash)
     VALUE key, value;
@@ -1492,7 +1528,7 @@ rb_hash_invert(hash)
 {
     VALUE h = rb_hash_new();
 
-    st_foreach(RHASH(hash)->tbl, rb_hash_invert_i, h);
+    rb_hash_foreach(hash, rb_hash_invert_i, h);
     return h;
 }
 
@@ -1521,8 +1557,10 @@ rb_hash_update_block_i(key, value, hash)
 
 /*
  *  call-seq:
- *     hsh.merge!(other_hash)    => hsh
- *     hsh.update(other_hash)    => hsh
+ *     hsh.merge!(other_hash)                                 => hsh
+ *     hsh.update(other_hash)                                 => hsh
+ *     hsh.merge!(other_hash){|key, oldval, newval| block}    => hsh
+ *     hsh.update(other_hash){|key, oldval, newval| block}    => hsh
  *  
  *  Adds the contents of <i>other_hash</i> to <i>hsh</i>, overwriting
  *  entries with duplicate keys with those from <i>other_hash</i>.
@@ -1538,17 +1576,18 @@ rb_hash_update(hash1, hash2)
 {
     hash2 = to_hash(hash2);
     if (rb_block_given_p()) {
-	st_foreach(RHASH(hash2)->tbl, rb_hash_update_block_i, hash1);
+	rb_hash_foreach(hash2, rb_hash_update_block_i, hash1);
     }
     else {
-	st_foreach(RHASH(hash2)->tbl, rb_hash_update_i, hash1);
+	rb_hash_foreach(hash2, rb_hash_update_i, hash1);
     }
     return hash1;
 }
 
 /*
  *  call-seq:
- *     hsh.merge(other_hash) -> a_hash
+ *     hsh.merge(other_hash)                              -> a_hash
+ *     hsh.merge(other_hash){|key, oldval, newval| block} -> a_hash
  *  
  *  Returns a new hash containing the contents of <i>other_hash</i> and
  *  the contents of <i>hsh</i>, overwriting entries in <i>hsh</i> with
@@ -1849,7 +1888,7 @@ env_aset(obj, nm, val)
     char *name, *value;
 
     if (rb_safe_level() >= 4) {
-	rb_raise(rb_eSecurityError, "cannot change environment variable");
+	rb_raise(rb_eSecurityError, "can't change environment variable");
     }
 
     if (NIL_P(val)) {
@@ -2328,7 +2367,7 @@ env_replace(env, hash)
 
     if (env == hash) return env;
     hash = to_hash(hash);
-    st_foreach(RHASH(hash)->tbl, env_replace_i, keys);
+    rb_hash_foreach(hash, env_replace_i, keys);
 
     for (i=0; i<RARRAY(keys)->len; i++) {
 	env_delete(env, RARRAY(keys)->ptr[i]);
@@ -2355,7 +2394,7 @@ env_update(env, hash)
 {
     if (env == hash) return env;
     hash = to_hash(hash);
-    st_foreach(RHASH(hash)->tbl, env_update_i, 0);
+    rb_hash_foreach(hash, env_update_i, 0);
     return env;
 }
 
@@ -2395,7 +2434,6 @@ Init_Hash()
     rb_define_method(rb_cHash,"inspect", rb_hash_inspect, 0);
 
     rb_define_method(rb_cHash,"==", rb_hash_equal, 1);
-    rb_define_method(rb_cHash,"eql?", rb_hash_eql, 1);
     rb_define_method(rb_cHash,"[]", rb_hash_aref, 1);
     rb_define_method(rb_cHash,"fetch", rb_hash_fetch, -1);
     rb_define_method(rb_cHash,"[]=", rb_hash_aset, 2);
