@@ -430,6 +430,15 @@ module DRb
     end
   end
 
+  class DRbRemoteError < DRbError
+    def initialize(error)
+      @reason = error.class.to_s
+      super("#{error.message} (#{error.class})")
+      set_backtrace(error.backtrace)
+    end
+    attr_reader :reason
+  end
+
   # Class wrapping a marshalled object whose type is unknown locally.
   #
   # If an object is returned by a method invoked over drb, but the
@@ -452,7 +461,7 @@ module DRb
     # when the unmarshalling failed.  It is used to determine the
     # name of the unmarshalled object.
     def initialize(err, buf)
-      case err
+      case err.to_s
       when /uninitialized constant (\S+)/
 	@name = $1
       when /undefined class\/module (\S+)/
@@ -539,12 +548,12 @@ module DRb
       @argc_limit = config[:argc_limit]
     end
 
-    def dump(obj)  # :nodoc:
-      obj = DRbObject.new(obj) if obj.kind_of? DRbUndumped
+    def dump(obj, error=false)  # :nodoc:
+      obj = make_proxy(obj, error) if obj.kind_of? DRbUndumped
       begin
 	str = Marshal::dump(obj)
       rescue
-	str = Marshal::dump(DRbObject.new(obj))
+	str = Marshal::dump(make_proxy(obj, error))
       end
       [str.size].pack('N') + str
     end
@@ -564,12 +573,21 @@ module DRb
       rescue
         raise(DRbConnError, $!.message, $!.backtrace)
       end
-      raise(DRbConnError, 'connection closed') if sz.nil?
+      raise(DRbConnError, 'connection closed') if str.nil?
       raise(DRbConnError, 'premature marshal format(can\'t read)') if str.size < sz
-      begin
-	Marshal::load(str)
-      rescue NameError, ArgumentError
-	DRbUnknown.new($!, str)
+      Thread.exclusive do
+        begin
+          save = Thread.current[:drb_untaint]
+          Thread.current[:drb_untaint] = []
+          Marshal::load(str)
+        rescue NameError, ArgumentError
+          DRbUnknown.new($!, str)
+        ensure
+          Thread.current[:drb_untaint].each do |x|
+            x.untaint
+          end
+          Thread.current[:drb_untaint] = save
+        end
       end
     end
 
@@ -602,7 +620,7 @@ module DRb
     end
 
     def send_reply(stream, succ, result)  # :nodoc:
-      stream.write(dump(succ) + dump(result))
+      stream.write(dump(succ) + dump(result, !succ))
     rescue
       raise(DRbConnError, $!.message, $!.backtrace)
     end
@@ -612,6 +630,15 @@ module DRb
       result = load(stream)
       [succ, result]
     end
+
+    private
+    def make_proxy(obj, error=false)
+      if error
+        DRbRemoteError.new(obj)
+      else
+        DRbObject.new(obj)
+      end
+    end
   end
 
   # Module managing the underlying network protocol(s) used by drb.
@@ -797,14 +824,40 @@ module DRb
       self.new(uri, soc, config)
     end
 
+    def self.getservername
+      host = Socket::gethostname
+      begin
+        Socket::gethostbyname(host)[0]
+      rescue
+        host
+      end
+    end
+
+    def self.open_server_inaddr_any(host, port)
+      infos = Socket::getaddrinfo(host, nil, 
+                                  Socket::AF_UNSPEC,
+                                  Socket::SOCK_STREAM, 
+                                  0,
+                                  Socket::AI_PASSIVE)
+      family = infos.collect { |af, *_| af }.uniq
+      case family
+      when ['AF_INET']
+        return TCPServer.open('0.0.0.0', port)
+      when ['AF_INET6']
+        return TCPServer.open('::', port)
+      else
+        return TCPServer.open(port)
+      end
+    end
+
     # Open a server listening for connections at +uri+ using 
     # configuration +config+.
     def self.open_server(uri, config)
       uri = 'druby://:0' unless uri
       host, port, opt = parse_uri(uri)
       if host.size == 0
-	soc = TCPServer.open(port)
-	host = Socket.gethostname
+        host = getservername
+        soc = open_server_inaddr_any(host, port)
       else
 	soc = TCPServer.open(host, port)
       end
@@ -944,12 +997,22 @@ module DRb
     # created to act as a stub for the remote referenced object.
     def self._load(s)
       uri, ref = Marshal.load(s)
+      
       if DRb.here?(uri)
-	return DRb.to_obj(ref)
+	obj = DRb.to_obj(ref)
+        if ((! obj.tainted?) && Thread.current[:drb_untaint])
+          Thread.current[:drb_untaint].push(obj)
+        end
+        return obj
       end
 
-      it = self.new(nil)
-      it.reinit(uri, ref)
+      self.new_with(uri, ref)
+    end
+
+    def self.new_with(uri, ref)
+      it = self.allocate
+      it.instance_variable_set('@uri', uri)
+      it.instance_variable_set('@ref', ref)
       it
     end
 
@@ -983,12 +1046,6 @@ module DRb
       end
     end
 
-    # Reinitialise this object with the given +uri+ and +ref+
-    def reinit(uri, ref)
-      @uri = uri
-      @ref = ref
-    end
-
     # Get the URI of the remote object.
     def __drburi 
       @uri
@@ -1001,7 +1058,17 @@ module DRb
 
     undef :to_s
     undef :to_a if respond_to?(:to_a)
-    undef :respond_to?
+
+    def respond_to?(msg_id, priv=false)
+      case msg_id
+      when :_dump
+        true
+      when :marshal_dump
+        false
+      else
+        method_missing(:respond_to?, msg_id, priv)
+      end
+    end
 
     # Routes method calls to the referenced object.
     def method_missing(msg_id, *a, &b)
@@ -1011,27 +1078,47 @@ module DRb
 	return obj.__send__(msg_id, *a, &b) 
       end
 
-      succ, result = DRbConn.open(@uri) do |conn|
-	conn.send_message(self, msg_id, a, b)
+      succ, result = self.class.with_friend(@uri) do
+        DRbConn.open(@uri) do |conn|
+          conn.send_message(self, msg_id, a, b)
+        end
       end
-      return result if succ
-      unless DRbUnknown === result
-	prefix = "(#{@uri}) "
-	bt = []
-	result.backtrace.each do |x|
-	  break if /`__send__'$/ =~ x 
-	  if /^\(druby:\/\// =~ x
-	    bt.push(x)
-	  else
-	    bt.push(prefix + x)
-	  end
-	end
-	raise result, result.message, bt + caller
+
+      if succ
+        return result
+      elsif DRbUnknown === result
+        raise result
       else
-	raise result
+        bt = self.class.prepare_backtrace(@uri, result)
+        raise result, result.message, bt + caller
       end
     end
 
+    def self.with_friend(uri)
+      friend = DRb.fetch_server(uri)
+      return yield() unless friend
+      
+      save = Thread.current['DRb']
+      Thread.current['DRb'] = { 'server' => friend }
+      return yield
+    ensure
+      Thread.current['DRb'] = save if friend
+    end
+
+    def self.prepare_backtrace(uri, result)
+      prefix = "(#{uri}) "
+      bt = []
+      result.backtrace.each do |x|
+        break if /`__send__'$/ =~ x 
+        if /^\(druby:\/\// =~ x
+          bt.push(x)
+        else
+          bt.push(prefix + x)
+        end
+      end
+      bt
+    end
+
     def pretty_print(q)   # :nodoc:
       q.pp_object(self)
     end
@@ -1132,6 +1219,7 @@ module DRb
     @@argc_limit = 256
     @@load_limit = 256 * 102400
     @@verbose = false
+    @@safe_level = 0
 
     # Set the default value for the :argc_limit option.
     #
@@ -1161,6 +1249,10 @@ module DRb
       @@idconv = idconv
     end
 
+    def self.default_safe_level(level)
+      @@safe_level = level
+    end
+
     # Set the default value of the :verbose option.
     #
     # See #new().  The initial default value is false.
@@ -1179,7 +1271,8 @@ module DRb
 	:verbose => @@verbose,
 	:tcp_acl => @@acl,
 	:load_limit => @@load_limit,
-	:argc_limit => @@argc_limit
+	:argc_limit => @@argc_limit,
+        :safe_level => @@safe_level
       }
       default_config.update(hash)
     end
@@ -1244,13 +1337,12 @@ module DRb
 
       @front = front
       @idconv = @config[:idconv]
+      @safe_level = @config[:safe_level]
 
       @grp = ThreadGroup.new
       @thread = run
 
-      Thread.exclusive do
-	DRb.primary_server = self unless DRb.primary_server
-      end
+      DRb.regist_server(self)
     end
 
     # The URI of this DRbServer.
@@ -1272,6 +1364,8 @@ module DRb
     # The configuration of this DRbServer
     attr_reader :config
 
+    attr_reader :safe_level
+
     # Set whether to operate in verbose mode.
     #
     # In verbose mode, failed calls are logged to stdout.
@@ -1289,7 +1383,12 @@ module DRb
 
     # Stop this server.
     def stop_service
-      @thread.kill
+      DRb.remove_server(self)
+      if  Thread.current['DRb'] && Thread.current['DRb']['server'] == self
+        Thread.current['DRb']['stop_service'] = true
+      else
+        @thread.kill
+      end
     end
 
     # Convert a dRuby reference to the local object it refers to.
@@ -1382,6 +1481,7 @@ module DRb
     class InvokeMethod  # :nodoc:
       def initialize(drb_server, client)
 	@drb_server = drb_server
+        @safe_level = drb_server.safe_level
 	@client = client
       end
 
@@ -1389,10 +1489,28 @@ module DRb
 	@result = nil
 	@succ = false
 	setup_message
-        if @block
-          @result = perform_with_block
+
+        if $SAFE < @safe_level
+          info = Thread.current['DRb']
+          if @block
+            @result = Thread.new {
+              Thread.current['DRb'] = info
+              $SAFE = @safe_level
+              perform_with_block
+            }.value
+          else
+            @result = Thread.new { 
+              Thread.current['DRb'] = info
+              $SAFE = @safe_level
+              perform_without_block
+            }.value
+          end
         else
-          @result = perform_without_block
+          if @block
+            @result = perform_with_block
+          else
+            @result = perform_without_block
+          end
         end
 	@succ = true
 	if @msg_id == :to_ary && @result.class == Array
@@ -1474,10 +1592,11 @@ module DRb
 	    end
 	    client.send_reply(succ, result) rescue nil
 	  ensure
-	    unless succ
-	      client.close
-	      break
-	    end
+            client.close unless succ
+            if Thread.current['DRb']['stop_service']
+              Thread.new { stop_service }
+            end
+            break unless succ
 	  end
 	end
       end
@@ -1615,6 +1734,25 @@ module DRb
     DRbServer.default_acl(acl)
   end
   module_function :install_acl
+
+  @server = {}
+  def regist_server(server)
+    @server[server.uri] = server
+    Thread.exclusive do
+      @primary_server = server unless @primary_server
+    end
+  end
+  module_function :regist_server
+
+  def remove_server(server)
+    @server.delete(server.uri)
+  end
+  module_function :remove_server
+  
+  def fetch_server(uri)
+    @server[uri]
+  end
+  module_function :fetch_server
 end
 
 DRbObject = DRb::DRbObject
