@@ -1,7 +1,7 @@
 # = uri/common.rb
 #
 # Author:: Akira Yamada <akira@ruby-lang.org>
-# Revision:: $Id: common.rb,v 1.11.2.3 2004/07/05 14:06:09 akr Exp $
+# Revision:: $Id: common.rb,v 1.11.2.5 2005/01/16 15:11:39 eban Exp $
 # License:: 
 #   You can redistribute it and/or modify it under the same term as Ruby.
 #
@@ -88,7 +88,7 @@ module URI
       # host          = hostname | IPv4address | IPv6reference (RFC 2732)
       HOST = "(?:#{HOSTNAME}|#{IPV4ADDR}|#{IPV6REF})"
       # port          = *digit
-      PORT = "\d*"
+      PORT = '\d*'
       # hostport      = host [ ":" port ]
       HOSTPORT = "#{HOST}(?::#{PORT})?"
 
@@ -214,7 +214,7 @@ module URI
     # :startdoc:
   end # REGEXP
 
-  module Util
+  module Util # :nodoc:
     def make_components_hash(klass, array_hash)
       tmp = {}
       if array_hash.kind_of?(Array) &&
@@ -260,7 +260,7 @@ module URI
     #   String to replaces in.
     # +unsafe+::
     #   Regexp that matches all symbols that must be replaced with codes.
-    #   By default uses <tt>REGEXP::SAFE</tt>.
+    #   By default uses <tt>REGEXP::UNSAFE</tt>.
     #
     # == Description
     #
