@@ -2,8 +2,8 @@
 
   object.c -
 
-  $Author: ocean $
-  $Date: 2004/07/15 11:42:51 $
+  $Author: matz $
+  $Date: 2005/05/11 01:11:48 $
   created at: Thu Jul 15 12:01:24 JST 1993
 
   Copyright (C) 1993-2003 Yukihiro Matsumoto
@@ -149,7 +149,7 @@ VALUE
 rb_obj_id_obsolete(obj)
     VALUE obj;
 {
-    rb_warning("Object#id will be deprecated; use Object#object_id");
+    rb_warn("Object#id will be deprecated; use Object#object_id");
     return rb_obj_id(obj);
 }
 
@@ -404,7 +404,7 @@ static VALUE
 inspect_obj(obj, str)
     VALUE obj, str;
 {
-    st_foreach(ROBJECT(obj)->iv_tbl, inspect_i, str);
+    st_foreach_safe(ROBJECT(obj)->iv_tbl, inspect_i, str);
     rb_str_cat2(str, ">");
     RSTRING(str)->ptr[0] = '#';
     OBJ_INFECT(str, obj);
@@ -529,7 +529,7 @@ rb_obj_is_kind_of(obj, c)
 
 
 /*
- * Document-method: singleton-method-added
+ * Document-method: singleton_method_added
  *
  *  call-seq:
  *     singleton_method_added(symbol)
@@ -555,7 +555,7 @@ rb_obj_is_kind_of(obj, c)
  */
 
 /*
- * Document-method: singleton-method-removed
+ * Document-method: singleton_method_removed
  *
  *  call-seq:
  *     singleton_method_removed(symbol)
@@ -583,7 +583,7 @@ rb_obj_is_kind_of(obj, c)
  */
 
 /*
- * Document-method: singleton-method-undefined
+ * Document-method: singleton_method_undefined
  *
  *  call-seq:
  *     singleton_method_undefined(symbol)
@@ -607,6 +607,31 @@ rb_obj_is_kind_of(obj, c)
  */
 
 
+/*
+ * Document-method: included
+ *
+ * call-seq:
+ *    included( othermod )
+ *
+ * Callback invoked whenever the receiver is included in another
+ * module or class. This should be used in preference to
+ * <tt>Module.append_features</tt> if your code wants to perform some
+ * action when a module is included in another.
+ *
+ *        module A
+ *          def A.included(mod)
+ *            puts "#{self} included in #{mod}"
+ *          end
+ *        end
+ *        module Enumerable
+ *          include A
+ *        end
+ */
+
+
+/*
+ * Not documented
+ */
 
 static VALUE
 rb_obj_dummy()
@@ -809,7 +834,7 @@ nil_to_a(obj)
 
 /*
  *  call-seq:
- *    nil.to_s  => "nil"
+ *    nil.inspect  => "nil"
  *
  *  Always returns the string "nil".
  */
@@ -1051,8 +1076,8 @@ rb_false(obj)
  */
 
 static VALUE
-rb_obj_pattern_match(obj)
-    VALUE obj;
+rb_obj_pattern_match(obj1, obj2)
+    VALUE obj1, obj2;
 {
     return Qfalse;
 }
@@ -1093,7 +1118,6 @@ rb_obj_pattern_match(obj)
 /*
  *  call-seq:
  *     sym.to_i      => fixnum
- *     sym.to_int    => fixnum
  *  
  *  Returns an integer that is unique for each symbol within a
  *  particular execution of a program.
@@ -1112,6 +1136,17 @@ sym_to_i(sym)
 }
 
 
+/* :nodoc: */
+
+static VALUE
+sym_to_int(sym)
+    VALUE sym;
+{
+    rb_warning("treating Symbol as an integer");
+    return sym_to_i(sym);
+}
+
+
 /*
  *  call-seq:
  *     sym.inspect    => string
@@ -1283,8 +1318,8 @@ rb_mod_eqq(mod, arg)
  *
  */
 
-static VALUE
-rb_mod_le(mod, arg)
+VALUE
+rb_class_inherited_p(mod, arg)
     VALUE mod, arg;
 {
     VALUE start = mod;
@@ -1333,7 +1368,7 @@ rb_mod_lt(mod, arg)
     VALUE mod, arg;
 {
     if (mod == arg) return Qfalse;
-    return rb_mod_le(mod, arg);
+    return rb_class_inherited_p(mod, arg);
 }
 
 
@@ -1361,7 +1396,7 @@ rb_mod_ge(mod, arg)
 	rb_raise(rb_eTypeError, "compared with non class/module");
     }
 
-    return rb_mod_le(arg, mod);
+    return rb_class_inherited_p(arg, mod);
 }
 
 /*
@@ -1409,7 +1444,7 @@ rb_mod_cmp(mod, arg)
 	return Qnil;
     }
 
-    cmp = rb_mod_le(mod, arg);
+    cmp = rb_class_inherited_p(mod, arg);
     if (NIL_P(cmp)) return Qnil;
     if (cmp) {
 	return INT2FIX(-1);
@@ -1589,6 +1624,9 @@ rb_class_superclass(klass)
     if (!super) {
 	rb_raise(rb_eTypeError, "uninitialized class");
     }
+    if (FL_TEST(klass, FL_SINGLETON)) {
+	super = RBASIC(klass)->klass;
+    }
     while (TYPE(super) == T_ICLASS) {
 	super = RCLASS(super)->super;
     }
@@ -1605,6 +1643,9 @@ str_to_id(str)
     if (!RSTRING(str)->ptr || RSTRING(str)->len == 0) {
 	rb_raise(rb_eArgError, "empty symbol string");
     }
+    if (RSTRING(str)->len != strlen(RSTRING(str)->ptr)) {
+	rb_warn("Symbols should not contain NUL (\\0)");
+    }
     return rb_intern(RSTRING(str)->ptr);
 }
 
@@ -1971,7 +2012,8 @@ rb_obj_ivar_get(obj, iv)
  *  
  *  Sets the instance variable names by <i>symbol</i> to
  *  <i>object</i>, thereby frustrating the efforts of the class's
- *  author to attempt to provide proper encapsulation.
+ *  author to attempt to provide proper encapsulation. The variable
+ *  did not have to exist prior to this call.
  *     
  *     class Fred
  *       def initialize(p1, p2)
@@ -1980,7 +2022,8 @@ rb_obj_ivar_get(obj, iv)
  *     end
  *     fred = Fred.new('cat', 99)
  *     fred.instance_variable_set(:@a, 'dog')   #=> "dog"
- *     fred.inspect                             #=> "#<Fred:0x401b3da8 @a=\"dog\", @b=99>"
+ *     fred.instance_variable_set(:@c, 'cat')   #=> "cat"
+ *     fred.inspect                             #=> "#<Fred:0x401b3da8 @a=\"dog\", @b=99, @c=\"cat\">"
  */
 
 static VALUE
@@ -2006,7 +2049,7 @@ convert_type(val, tname, method, raise)
     m = rb_intern(method);
     if (!rb_respond_to(val, m)) {
 	if (raise) {
-	    rb_raise(rb_eTypeError, "cannot convert %s into %s",
+	    rb_raise(rb_eTypeError, "can't convert %s into %s",
 		     NIL_P(val) ? "nil" :
 		     val == Qtrue ? "true" :
 		     val == Qfalse ? "false" :
@@ -2081,6 +2124,8 @@ VALUE
 rb_Integer(val)
     VALUE val;
 {
+    VALUE tmp;
+
     switch (TYPE(val)) {
       case T_FLOAT:
 	if (RFLOAT(val)->value <= (double)FIXNUM_MAX
@@ -2099,10 +2144,11 @@ rb_Integer(val)
       default:
 	break;
     }
-    if (rb_respond_to(val, rb_intern("to_int"))) {
-	return rb_to_integer(val, "to_int");
+    tmp = convert_type(val, "Integer", "to_int", Qfalse);
+    if (NIL_P(tmp)) {
+	return rb_to_integer(val, "to_i");
     }
-    return rb_to_integer(val, "to_i");
+    return tmp;
 }
 
 /*
@@ -2242,7 +2288,7 @@ rb_Float(val)
 	return rb_float_new(rb_str_to_dbl(val, Qtrue));
 
       case T_NIL:
-	rb_raise(rb_eTypeError, "cannot convert nil into Float");
+	rb_raise(rb_eTypeError, "can't convert nil into Float");
 	break;
 
       default:
@@ -2344,18 +2390,10 @@ rb_Array(val)
     VALUE val;
 {
     VALUE tmp = rb_check_array_type(val);
-    ID to_a;
 
     if (NIL_P(tmp)) {
-	to_a = rb_intern("to_a");
-	if (rb_respond_to(val, to_a)) {
-	    val = rb_funcall(val, to_a, 0);
-	    if (TYPE(val) != T_ARRAY) {
-		rb_raise(rb_eTypeError, "`to_a' did not return Array");
-	    }
-	    return val;
-	}
-	else {
+	tmp = rb_check_convert_type(val, T_ARRAY, "Array", "to_a");
+	if (NIL_P(tmp)) {
 	    return rb_ary_new3(1, val);
 	}
     }
@@ -2433,7 +2471,7 @@ VALUE ruby_top_self;
  *     Creating a new Name
  *     
  *  Classes, modules, and objects are interrelated. In the diagram
- *  that follows, the arrows represent inheritance, and the
+ *  that follows, the vertical arrows represent inheritance, and the
  *  parentheses meta-classes. All metaclasses are instances 
  *  of the class `Class'.
  *
@@ -2579,18 +2617,19 @@ Init_Object()
     rb_undef_method(CLASS_OF(rb_cSymbol), "new");
 
     rb_define_method(rb_cSymbol, "to_i", sym_to_i, 0);
-    rb_define_method(rb_cSymbol, "to_int", sym_to_i, 0);
+    rb_define_method(rb_cSymbol, "to_int", sym_to_int, 0);
     rb_define_method(rb_cSymbol, "inspect", sym_inspect, 0);
     rb_define_method(rb_cSymbol, "to_s", sym_to_s, 0);
     rb_define_method(rb_cSymbol, "id2name", sym_to_s, 0);
     rb_define_method(rb_cSymbol, "to_sym", sym_to_sym, 0);
+    rb_define_method(rb_cSymbol, "===", rb_obj_equal, 1); 
 
     rb_define_method(rb_cModule, "freeze", rb_mod_freeze, 0);
     rb_define_method(rb_cModule, "===", rb_mod_eqq, 1);
     rb_define_method(rb_cModule, "==", rb_obj_equal, 1);
     rb_define_method(rb_cModule, "<=>",  rb_mod_cmp, 1);
     rb_define_method(rb_cModule, "<",  rb_mod_lt, 1);
-    rb_define_method(rb_cModule, "<=", rb_mod_le, 1);
+    rb_define_method(rb_cModule, "<=", rb_class_inherited_p, 1);
     rb_define_method(rb_cModule, ">",  rb_mod_gt, 1);
     rb_define_method(rb_cModule, ">=", rb_mod_ge, 1);
     rb_define_method(rb_cModule, "initialize_copy", rb_mod_init_copy, 1);
