@@ -2,8 +2,8 @@
 
   process.c -
 
-  $Author: nobu $
-  $Date: 2004/04/15 15:02:12 $
+  $Author: matz $
+  $Date: 2005/02/28 02:45:19 $
   created at: Tue Aug 10 14:30:50 JST 1993
 
   Copyright (C) 1993-2003 Yukihiro Matsumoto
@@ -91,6 +91,11 @@ static VALUE S_Tms;
 #define HAVE_44BSD_SETGID 1
 #endif
 
+#ifdef __NetBSD__  
+#undef HAVE_SETRUID 
+#undef HAVE_SETRGID
+#endif
+
 #if defined(__MacOS_X__) || defined(__bsdi__)
 #define BROKEN_SETREUID 1
 #define BROKEN_SETREGID 1
@@ -119,6 +124,7 @@ static VALUE S_Tms;
 static VALUE
 get_pid()
 {
+    rb_secure(2);
     return INT2FIX(getpid());
 }
 
@@ -142,6 +148,7 @@ get_pid()
 static VALUE
 get_ppid()
 {
+    rb_secure(2);
 #ifdef _WIN32
     return INT2FIX(0);
 #else
@@ -719,6 +726,7 @@ proc_wait(argc, argv)
     VALUE vpid, vflags;
     int pid, flags, status;
 
+    rb_secure(2);
     flags = 0;
     rb_scan_args(argc, argv, "02", &vpid, &vflags);
     if (argc == 0) {
@@ -793,6 +801,7 @@ proc_waitall()
     VALUE result;
     int pid, status;
 
+    rb_secure(2);
     result = rb_ary_new();
 #ifdef NO_WAITPID
     if (pid_tbl) {
@@ -894,6 +903,7 @@ static VALUE
 proc_detach(obj, pid)
     VALUE pid;
 {
+    rb_secure(2);
     return rb_detach_process(NUM2INT(pid));
 }
 
@@ -1214,8 +1224,8 @@ rb_f_exec(argc, argv)
 	    rb_raise(rb_eArgError, "wrong first argument");
 	}
 	prog = RARRAY(tmp)->ptr[0];
-	SafeStringValue(prog);
 	argv[0] = RARRAY(tmp)->ptr[1];
+	SafeStringValue(prog);
     }
     if (argc == 1 && prog == 0) {
 	VALUE cmd = argv[0];
@@ -1233,9 +1243,20 @@ rb_f_exec(argc, argv)
 
 /*
  *  call-seq:
+ *     Kernel.fork  [{ block }]   => fixnum or nil
  *     Process.fork [{ block }]   => fixnum or nil
- *  
- *  See <code>Kernel::fork</code>.
+ *
+ *  Creates a subprocess. If a block is specified, that block is run
+ *  in the subprocess, and the subprocess terminates with a status of
+ *  zero. Otherwise, the +fork+ call returns twice, once in
+ *  the parent, returning the process ID of the child, and once in
+ *  the child, returning _nil_. The child process can exit using
+ *  <code>Kernel.exit!</code> to avoid running any
+ *  <code>at_exit</code> functions. The parent process should 
+ *  use <code>Process.wait</code> to collect the termination statuses 
+ *  of its children or use <code>Process.detach</code> to register
+ *  disinterest in their status; otherwise, the operating system
+ *  may accumulate zombie processes.
  */
 
 static VALUE
@@ -1246,6 +1267,12 @@ rb_f_fork(obj)
     int pid;
 
     rb_secure(2);
+
+#ifndef __VMS
+    fflush(stdout);
+    fflush(stderr);
+#endif
+
     switch (pid = fork()) {
       case 0:
 #ifdef linux
@@ -1315,6 +1342,10 @@ rb_f_exit_bang(argc, argv, obj)
     return Qnil;		/* not reached */
 }
 
+#if defined(sun)
+#define signal(a,b) sigset(a,b)
+#endif
+
 void
 rb_syswait(pid)
     int pid;
@@ -1427,9 +1458,7 @@ rb_f_system(argc, argv)
     else {
 	status = proc_spawn_n(argc, argv, prog);
     }
-#if defined(_WIN32)
-    last_status_set(status, 0);
-#else
+#if !defined(_WIN32)
     last_status_set(status == -1 ? 127 : status, 0);
 #endif
 #elif defined(__VMS)
@@ -1510,13 +1539,13 @@ rb_f_system(argc, argv)
 
 /*
  *  call-seq:
- *     sleep(duration=0)    => fixnum
+ *     sleep([duration])    => fixnum
  *  
  *  Suspends the current thread for _duration_ seconds (which may be
  *  any number, including a +Float+ with fractional seconds). Returns the actual
  *  number of seconds slept (rounded), which may be less than that asked
  *  for if the thread was interrupted by a +SIGALRM+, or if
- *  another thread calls <code>Thread#run</code>. An argument of zero
+ *  another thread calls <code>Thread#run</code>. Zero arguments
  *  causes +sleep+ to sleep forever.
  *     
  *     Time.new    #=> Wed Apr 09 08:56:32 CDT 2003
@@ -1566,6 +1595,7 @@ proc_getpgrp()
 {
     int pgrp;
 
+    rb_secure(2);
 #if defined(HAVE_GETPGRP) && defined(GETPGRP_VOID)
     pgrp = getpgrp();
     if (pgrp < 0) rb_sys_fail(0);
@@ -1593,12 +1623,13 @@ proc_getpgrp()
 static VALUE
 proc_setpgrp()
 {
+    rb_secure(2);
   /* check for posix setpgid() first; this matches the posix */
   /* getpgrp() above.  It appears that configure will set SETPGRP_VOID */
   /* even though setpgrp(0,0) would be prefered. The posix call avoids */
   /* this confusion. */
 #ifdef HAVE_SETPGID
-  if (setpgid(0,0) < 0) rb_sys_fail(0);
+    if (setpgid(0,0) < 0) rb_sys_fail(0);
 #elif defined(HAVE_SETPGRP) && defined(SETPGRP_VOID)
     if (setpgrp() < 0) rb_sys_fail(0);
 #else
@@ -1623,8 +1654,10 @@ proc_getpgid(obj, pid)
     VALUE obj, pid;
 {
 #if defined(HAVE_GETPGID) && !defined(__CHECKER__)
-    int i = getpgid(NUM2INT(pid));
+    int i;
 
+    rb_secure(2);
+    i = getpgid(NUM2INT(pid));
     if (i < 0) rb_sys_fail(0);
     return INT2NUM(i);
 #else
@@ -1732,6 +1765,7 @@ proc_getpriority(obj, which, who)
 #ifdef HAVE_GETPRIORITY
     int prio, iwhich, iwho;
 
+    rb_secure(2);
     iwhich = NUM2INT(which);
     iwho   = NUM2INT(who);
 
@@ -1783,7 +1817,7 @@ check_uid_switch()
 {
     rb_secure(2);
     if (under_uid_switch) {
-	rb_raise(rb_eRuntimeError, "can't handle UID during evaluating the block given to the Process::UID.switch method");
+	rb_raise(rb_eRuntimeError, "can't handle UID while evaluating block given to Process::UID.switch method");
     }
 }
 
@@ -1793,7 +1827,7 @@ check_gid_switch()
 {
     rb_secure(2);
     if (under_gid_switch) {
-	rb_raise(rb_eRuntimeError, "can't handle GID during evaluating the block given to the Process::UID.switch method");
+	rb_raise(rb_eRuntimeError, "can't handle GID while evaluating block given to Process::UID.switch method");
     }
 }
 
@@ -2290,6 +2324,7 @@ p_sys_issetugid(obj)
     VALUE obj;
 {
 #if defined HAVE_ISSETUGID
+    rb_secure(2);
     if (issetugid()) {
 	return Qtrue;
     } else {
@@ -2429,7 +2464,7 @@ proc_setgroups(VALUE obj, VALUE ary)
 
     groups = ALLOCA_N(gid_t, ngroups);
 
-    for (i = 0; i < ngroups; i++) {
+    for (i = 0; i < ngroups && i < RARRAY(ary)->len; i++) {
         VALUE g = RARRAY(ary)->ptr[i];
 
 	if (FIXNUM_P(g)) {
@@ -2442,10 +2477,10 @@ proc_setgroups(VALUE obj, VALUE ary)
 		groups[i] = NUM2INT(g);
 	    }
 	    else {
-		gr = getgrnam(RSTRING(g)->ptr);
+		gr = getgrnam(RSTRING(tmp)->ptr);
 		if (gr == NULL)
 		    rb_raise(rb_eArgError, 
-			     "can't find group for %s", RSTRING(g)->ptr);
+			     "can't find group for %s", RSTRING(tmp)->ptr);
 		groups[i] = gr->gr_gid;
 	    }
 	}
@@ -3430,26 +3465,18 @@ Init_process()
     rb_define_module_function(rb_mProcGID, "rid", proc_getgid, 0);
     rb_define_module_function(rb_mProcUID, "eid", proc_geteuid, 0);
     rb_define_module_function(rb_mProcGID, "eid", proc_getegid, 0);
-    rb_define_module_function(rb_mProcUID, "change_privilege", 
-			      p_uid_change_privilege, 1);
-    rb_define_module_function(rb_mProcGID, "change_privilege", 
-			      p_gid_change_privilege, 1);
-    rb_define_module_function(rb_mProcUID, "grant_privilege", 
-			      p_uid_grant_privilege, 1);
-    rb_define_module_function(rb_mProcGID, "grant_privilege", 
-			      p_gid_grant_privilege, 1);
+    rb_define_module_function(rb_mProcUID, "change_privilege", p_uid_change_privilege, 1);
+    rb_define_module_function(rb_mProcGID, "change_privilege", p_gid_change_privilege, 1);
+    rb_define_module_function(rb_mProcUID, "grant_privilege", p_uid_grant_privilege, 1);
+    rb_define_module_function(rb_mProcGID, "grant_privilege", p_gid_grant_privilege, 1);
     rb_define_alias(rb_mProcUID, "eid=", "grant_privilege");
     rb_define_alias(rb_mProcGID, "eid=", "grant_privilege");
     rb_define_module_function(rb_mProcUID, "re_exchange", p_uid_exchange, 0);
     rb_define_module_function(rb_mProcGID, "re_exchange", p_gid_exchange, 0);
-    rb_define_module_function(rb_mProcUID, "re_exchangeable?", 
-			      p_uid_exchangeable, 0);
-    rb_define_module_function(rb_mProcGID, "re_exchangeable?", 
-			      p_gid_exchangeable, 0);
-    rb_define_module_function(rb_mProcUID, "sid_available?", 
-			      p_uid_have_saved_id, 0);
-    rb_define_module_function(rb_mProcGID, "sid_available?", 
-			      p_gid_have_saved_id, 0);
+    rb_define_module_function(rb_mProcUID, "re_exchangeable?", p_uid_exchangeable, 0);
+    rb_define_module_function(rb_mProcGID, "re_exchangeable?", p_gid_exchangeable, 0);
+    rb_define_module_function(rb_mProcUID, "sid_available?", p_uid_have_saved_id, 0);
+    rb_define_module_function(rb_mProcGID, "sid_available?", p_gid_have_saved_id, 0);
     rb_define_module_function(rb_mProcUID, "switch", p_uid_switch, 0);
     rb_define_module_function(rb_mProcGID, "switch", p_gid_switch, 0);
 
@@ -3469,15 +3496,10 @@ Init_process()
     rb_define_module_function(rb_mProcID_Syscall, "seteuid", p_sys_seteuid, 1);
     rb_define_module_function(rb_mProcID_Syscall, "setegid", p_sys_setegid, 1);
 
-    rb_define_module_function(rb_mProcID_Syscall, "setreuid", 
-			      p_sys_setreuid, 2);
-    rb_define_module_function(rb_mProcID_Syscall, "setregid", 
-			      p_sys_setregid, 2);
+    rb_define_module_function(rb_mProcID_Syscall, "setreuid", p_sys_setreuid, 2);
+    rb_define_module_function(rb_mProcID_Syscall, "setregid", p_sys_setregid, 2);
 
-    rb_define_module_function(rb_mProcID_Syscall, "setresuid", 
-			      p_sys_setresuid, 3);
-    rb_define_module_function(rb_mProcID_Syscall, "setresgid", 
-			      p_sys_setresgid, 3);
-    rb_define_module_function(rb_mProcID_Syscall, "issetugid", 
-			      p_sys_issetugid, 0);
+    rb_define_module_function(rb_mProcID_Syscall, "setresuid", p_sys_setresuid, 3);
+    rb_define_module_function(rb_mProcID_Syscall, "setresgid", p_sys_setresgid, 3);
+    rb_define_module_function(rb_mProcID_Syscall, "issetugid", p_sys_issetugid, 0);
 }
