@@ -2,8 +2,8 @@
 
   sprintf.c -
 
-  $Author: usa $
-  $Date: 2004/07/16 02:20:00 $
+  $Author: matz $
+  $Date: 2005/02/28 02:45:19 $
   created at: Fri Oct 15 10:39:26 JST 1993
 
   Copyright (C) 1993-2003 Yukihiro Matsumoto
@@ -25,31 +25,30 @@ remove_sign_bits(str, base)
     char *str;
     int base;
 {
-    char *s, *t, *end;
-    unsigned long len;
+    char *s, *t;
     
     s = t = str;
-    len = strlen(str);
-    end = str + len;
 
     if (base == 16) {
-	while (t<end && *t == 'f') {
+	while (*t == 'f') {
 	    t++;
 	}
     }
     else if (base == 8) {
 	if (*t == '3') t++;
-	while (t<end && *t == '7') {
+	while (*t == '7') {
 	    t++;
 	}
     }
     else if (base == 2) {
-	while (t<end && *t == '1') {
+	while (*t == '1') {
 	    t++;
 	}
     }
-    while (*t) *s++ = *t++;
-    *s = '\0';
+    if (t > s) {
+	while (*t) *s++ = *t++;
+	*s = '\0';
+    }
 
     return str;
 }
@@ -57,7 +56,7 @@ remove_sign_bits(str, base)
 static char
 sign_bits(base, p)
     int base;
-    char *p;
+    const char *p;
 {
     char c = '.';
 
@@ -108,7 +107,7 @@ sign_bits(base, p)
 	       (posarg = -1, GETNTHARG(n))))
 
 #define GETNTHARG(nth) \
-    ((nth >= argc) ? (rb_raise(rb_eArgError, "too few arguments."), 0) : argv[nth])
+    ((nth >= argc) ? (rb_raise(rb_eArgError, "too few arguments"), 0) : argv[nth])
 
 #define GETASTER(val) do { \
     t = p++; \
@@ -234,7 +233,8 @@ rb_f_sprintf(argc, argv)
     VALUE *argv;
 {
     VALUE fmt;
-    char *buf, *p, *end;
+    const char *p, *end;
+    char *buf;
     int blen, bsiz;
     VALUE result;
 
@@ -249,6 +249,7 @@ rb_f_sprintf(argc, argv)
     fmt = GETNTHARG(0);
     if (OBJ_TAINTED(fmt)) tainted = 1;
     StringValue(fmt);
+    fmt = rb_str_new4(fmt);
     p = RSTRING(fmt)->ptr;
     end = p + RSTRING(fmt)->len;
     blen = 0;
@@ -257,7 +258,7 @@ rb_f_sprintf(argc, argv)
     buf = RSTRING(result)->ptr;
 
     for (; p < end; p++) {
-	char *t;
+	const char *t;
 	int n;
 
 	for (t = p; t < end && *t != '%'; t++) ;
@@ -550,7 +551,7 @@ rb_f_sprintf(argc, argv)
 			    s += 2;
 			}
 		    }
-		    sprintf(fbuf, "%%l%c", *p);
+		    sprintf(fbuf, "%%l%c", *p == 'X' ? 'x' : *p);
 		    sprintf(s, fbuf, v);
 		    if (v < 0) {
 			char d = 0;
@@ -770,14 +771,12 @@ rb_f_sprintf(argc, argv)
     }
 
   sprint_exit:
-#if 0
     /* XXX - We cannot validiate the number of arguments because
      *       the format string may contain `n$'-style argument selector.
      */
-    if (RTEST(ruby_verbose) && nextarg < argc) {
+    if (RTEST(ruby_verbose) && posarg >= 0 && nextarg < argc) {
 	rb_raise(rb_eArgError, "too many arguments for format string");
     }
-#endif
     rb_str_resize(result, blen);
 
     if (tainted) OBJ_TAINT(result);
