@@ -3,7 +3,7 @@
   struct.c -
 
   $Author: matz $
-  $Date: 2004/04/05 07:45:14 $
+  $Date: 2005/05/11 01:11:48 $
   created at: Tue Mar 22 18:44:30 JST 1995
 
   Copyright (C) 1993-2003 Yukihiro Matsumoto
@@ -11,6 +11,7 @@
 **********************************************************************/
 
 #include "ruby.h"
+#include "env.h"
 
 VALUE rb_cStruct;
 
@@ -33,19 +34,41 @@ rb_struct_iv_get(c, name)
     }
 }
 
-static VALUE
-rb_struct_s_members(obj)
-    VALUE obj;
+VALUE
+rb_struct_s_members(klass)
+    VALUE klass;
 {
-    VALUE member, ary;
+    VALUE members = rb_struct_iv_get(klass, "__members__");
+
+    if (NIL_P(members)) {
+	rb_bug("non-initialized struct");
+    }
+    return members;
+}
+
+VALUE
+rb_struct_members(s)
+    VALUE s;
+{
+    VALUE members = rb_struct_s_members(rb_obj_class(s));
+
+    if (RSTRUCT(s)->len != RARRAY(members)->len) {
+	rb_raise(rb_eTypeError, "struct size differs (%d required %d given)",
+		 RARRAY(members)->len, RSTRUCT(s)->len);
+    }
+    return members;
+}
+
+static VALUE
+rb_struct_s_members_m(klass)
+    VALUE klass;
+{
+    VALUE members, ary;
     VALUE *p, *pend;
 
-    member = rb_struct_iv_get(obj, "__member__");
-    if (NIL_P(member)) {
-	rb_bug("uninitialized struct");
-    }
-    ary = rb_ary_new2(RARRAY(member)->len);
-    p = RARRAY(member)->ptr; pend = p + RARRAY(member)->len;
+    members = rb_struct_s_members(klass);
+    ary = rb_ary_new2(RARRAY(members)->len);
+    p = RARRAY(members)->ptr; pend = p + RARRAY(members)->len;
     while (p < pend) {
 	rb_ary_push(ary, rb_str_new2(rb_id2name(SYM2ID(*p))));
 	p++;
@@ -67,10 +90,10 @@ rb_struct_s_members(obj)
  */
 
 static VALUE
-rb_struct_members(obj)
+rb_struct_members_m(obj)
     VALUE obj;
 {
-    return rb_struct_s_members(rb_obj_class(obj));
+    return rb_struct_s_members_m(rb_obj_class(obj));
 }
 
 VALUE
@@ -78,16 +101,13 @@ rb_struct_getmember(obj, id)
     VALUE obj;
     ID id;
 {
-    VALUE member, slot;
+    VALUE members, slot;
     long i;
 
-    member = rb_struct_iv_get(rb_obj_class(obj), "__member__");
-    if (NIL_P(member)) {
-	rb_bug("uninitialized struct");
-    }
+    members = rb_struct_members(obj);
     slot = ID2SYM(id);
-    for (i=0; i<RARRAY(member)->len; i++) {
-	if (RARRAY(member)->ptr[i] == slot) {
+    for (i=0; i<RARRAY(members)->len; i++) {
+	if (RARRAY(members)->ptr[i] == slot) {
 	    return RSTRUCT(obj)->ptr[i];
 	}
     }
@@ -99,7 +119,7 @@ static VALUE
 rb_struct_ref(obj)
     VALUE obj;
 {
-    return rb_struct_getmember(obj, rb_frame_last_func());
+    return rb_struct_getmember(obj, ruby_frame->orig_func);
 }
 
 static VALUE rb_struct_ref0(obj) VALUE obj; {return RSTRUCT(obj)->ptr[0];}
@@ -139,33 +159,33 @@ static VALUE
 rb_struct_set(obj, val)
     VALUE obj, val;
 {
-    VALUE member, slot;
+    VALUE members, slot;
+    ID id;
     long i;
 
-    member = rb_struct_iv_get(rb_obj_class(obj), "__member__");
-    if (NIL_P(member)) {
-	rb_bug("uninitialized struct");
-    }
+    members = rb_struct_members(obj);
     rb_struct_modify(obj);
-    for (i=0; i<RARRAY(member)->len; i++) {
-	slot = RARRAY(member)->ptr[i];
-	if (rb_id_attrset(SYM2ID(slot)) == rb_frame_last_func()) {
+    id = ruby_frame->orig_func;
+    for (i=0; i<RARRAY(members)->len; i++) {
+	slot = RARRAY(members)->ptr[i];
+	if (rb_id_attrset(SYM2ID(slot)) == id) {
 	    return RSTRUCT(obj)->ptr[i] = val;
 	}
     }
-    rb_name_error(rb_frame_last_func(), "`%s' is not a struct member",
-		  rb_id2name(rb_frame_last_func()));
+    rb_name_error(ruby_frame->last_func, "`%s' is not a struct member",
+		  rb_id2name(id));
     return Qnil;		/* not reached */
 }
 
 static VALUE
-make_struct(name, member, klass)
-    VALUE name, member, klass;
+make_struct(name, members, klass)
+    VALUE name, members, klass;
 {
     VALUE nstr;
     ID id;
     long i;
 
+    OBJ_FREEZE(members);
     if (NIL_P(name)) {
 	nstr = rb_class_new(klass);
 	rb_make_metaclass(nstr, RBASIC(klass)->klass);
@@ -177,24 +197,30 @@ make_struct(name, member, klass)
 	if (!rb_is_const_id(id)) {
 	    rb_name_error(id, "identifier %s needs to be constant", cname);
 	}
-	nstr = rb_define_class_under(klass, cname, klass);
+	if (rb_const_defined_at(klass, id)) {
+	    rb_warn("redefining constant Struct::%s", cname);
+	    rb_mod_remove_const(klass, ID2SYM(id));
+	}
+	nstr = rb_define_class_under(klass, rb_id2name(id), klass);
     }
-    rb_iv_set(nstr, "__size__", LONG2NUM(RARRAY(member)->len));
-    rb_iv_set(nstr, "__member__", member);
+    rb_iv_set(nstr, "__size__", LONG2NUM(RARRAY(members)->len));
+    rb_iv_set(nstr, "__members__", members);
 
     rb_define_alloc_func(nstr, struct_alloc);
     rb_define_singleton_method(nstr, "new", rb_class_new_instance, -1);
     rb_define_singleton_method(nstr, "[]", rb_class_new_instance, -1);
-    rb_define_singleton_method(nstr, "members", rb_struct_s_members, 0);
-    for (i=0; i< RARRAY(member)->len; i++) {
-	ID id = SYM2ID(RARRAY(member)->ptr[i]);
-	if (i<10) {
-	    rb_define_method_id(nstr, id, ref_func[i], 0);
+    rb_define_singleton_method(nstr, "members", rb_struct_s_members_m, 0);
+    for (i=0; i< RARRAY(members)->len; i++) {
+	ID id = SYM2ID(RARRAY(members)->ptr[i]);
+	if (rb_is_local_id(id) || rb_is_const_id(id)) {
+	    if (i<10) {
+		rb_define_method_id(nstr, id, ref_func[i], 0);
+	    }
+	    else {
+		rb_define_method_id(nstr, id, rb_struct_ref, 0);
+	    }
+	    rb_define_method_id(nstr, rb_id_attrset(id), rb_struct_set, 1);
 	}
-	else {
-	    rb_define_method_id(nstr, id, rb_struct_ref, 0);
-	}
-	rb_define_method_id(nstr, rb_id_attrset(id), rb_struct_set, 1);
     }
 
     return nstr;
@@ -285,12 +311,19 @@ rb_struct_s_def(argc, argv, klass)
 	id = rb_to_id(RARRAY(rest)->ptr[i]);
 	RARRAY(rest)->ptr[i] = ID2SYM(id);
     }
-    if (!NIL_P(name) && TYPE(name) != T_STRING) {
-	id = rb_to_id(name);
-	rb_ary_unshift(rest, ID2SYM(id));
-	name = Qnil;
+    if (!NIL_P(name)) {
+	VALUE tmp = rb_check_string_type(name);
+
+	if (NIL_P(tmp)) {
+	    id = rb_to_id(name);
+	    rb_ary_unshift(rest, ID2SYM(id));
+	    name = Qnil;
+	}
     }
     st = make_struct(name, rest, klass);
+    if (rb_block_given_p()) {
+	rb_mod_module_eval(0, 0, st);
+    }
 
     return st;
 }
@@ -423,15 +456,12 @@ static VALUE
 rb_struct_each_pair(s)
     VALUE s;
 {
-    VALUE member;
+    VALUE members;
     long i;
 
-    member = rb_struct_iv_get(rb_obj_class(s), "__member__");
-    if (NIL_P(member)) {
-	rb_bug("non-initialized struct");
-    }
+    members = rb_struct_members(s);
     for (i=0; i<RSTRUCT(s)->len; i++) {
-	rb_yield_values(2, RARRAY(member)->ptr[i], RSTRUCT(s)->ptr[i]);
+	rb_yield_values(2, rb_ary_entry(members, i), RSTRUCT(s)->ptr[i]);
     }
     return s;
 }
@@ -441,30 +471,32 @@ inspect_struct(s)
     VALUE s;
 {
     char *cname = rb_class2name(rb_obj_class(s));
-    VALUE str, member;
+    VALUE str, members;
     long i;
 
-    member = rb_struct_iv_get(rb_obj_class(s), "__member__");
-    if (NIL_P(member)) {
-	rb_bug("non-initialized struct");
-    }
-
+    members = rb_struct_members(s);
     str = rb_str_buf_new2("#<struct ");
     rb_str_cat2(str, cname);
     rb_str_cat2(str, " ");
     for (i=0; i<RSTRUCT(s)->len; i++) {
-	VALUE str2, slot;
+	VALUE slot;
+	ID id;
 	char *p;
 
 	if (i > 0) {
 	    rb_str_cat2(str, ", ");
 	}
-	slot = RARRAY(member)->ptr[i];
-	p = rb_id2name(SYM2ID(slot));
-	rb_str_cat2(str, p);
+	slot = RARRAY(members)->ptr[i];
+	id = SYM2ID(slot);
+	if (rb_is_local_id(id) || rb_is_const_id(id)) {
+	    p = rb_id2name(id);
+	    rb_str_cat2(str, p);
+	}
+	else {
+	    rb_str_append(str, rb_inspect(slot));
+	}
 	rb_str_cat2(str, "=");
-	str2 = rb_inspect(RSTRUCT(s)->ptr[i]);
-	rb_str_append(str, str2);
+	rb_str_append(str, rb_inspect(RSTRUCT(s)->ptr[i]));
     }
     rb_str_cat2(str, ">");
     OBJ_INFECT(str, s);
@@ -514,6 +546,7 @@ rb_struct_to_a(s)
     return rb_ary_new4(RSTRUCT(s)->len, RSTRUCT(s)->ptr);
 }
 
+/* :nodoc: */
 static VALUE
 rb_struct_init_copy(copy, s)
     VALUE copy, s;
@@ -535,17 +568,13 @@ rb_struct_aref_id(s, id)
     VALUE s;
     ID id;
 {
-    VALUE member;
+    VALUE members;
     long i, len;
 
-    member = rb_struct_iv_get(rb_obj_class(s), "__member__");
-    if (NIL_P(member)) {
-	rb_bug("non-initialized struct");
-    }
-
-    len = RARRAY(member)->len;
+    members = rb_struct_members(s);
+    len = RARRAY(members)->len;
     for (i=0; i<len; i++) {
-	if (SYM2ID(RARRAY(member)->ptr[i]) == id) {
+	if (SYM2ID(RARRAY(members)->ptr[i]) == id) {
 	    return RSTRUCT(s)->ptr[i];
 	}
     }
@@ -598,18 +627,18 @@ rb_struct_aset_id(s, id, val)
     VALUE s, val;
     ID id;
 {
-    VALUE member;
+    VALUE members;
     long i, len;
 
-    member = rb_struct_iv_get(rb_obj_class(s), "__member__");
-    if (NIL_P(member)) {
-	rb_bug("non-initialized struct");
-    }
-
+    members = rb_struct_members(s);
     rb_struct_modify(s);
-    len = RARRAY(member)->len;
+    len = RARRAY(members)->len;
+    if (RSTRUCT(s)->len != RARRAY(members)->len) {
+	rb_raise(rb_eTypeError, "struct size differs (%d required %d given)",
+		 RARRAY(members)->len, RSTRUCT(s)->len);
+    }
     for (i=0; i<len; i++) {
-	if (SYM2ID(RARRAY(member)->ptr[i]) == id) {
+	if (SYM2ID(RARRAY(members)->ptr[i]) == id) {
 	    RSTRUCT(s)->ptr[i] = val;
 	    return val;
 	}
@@ -888,5 +917,5 @@ Init_Struct()
     rb_define_method(rb_cStruct, "select", rb_struct_select, -1);
     rb_define_method(rb_cStruct, "values_at", rb_struct_values_at, -1);
 
-    rb_define_method(rb_cStruct, "members", rb_struct_members, 0);
+    rb_define_method(rb_cStruct, "members", rb_struct_members_m, 0);
 }
