@@ -12,6 +12,7 @@
 
 #include "ruby.h"
 #include "rubysig.h"
+#include "dln.h"
 #include <fcntl.h>
 #include <process.h>
 #include <sys/stat.h>
@@ -88,12 +89,11 @@
 
 bool NtSyncProcess = TRUE;
 
-static struct ChildRecord *CreateChild(char *, char *, SECURITY_ATTRIBUTES *, HANDLE, HANDLE, HANDLE);
-static int make_cmdvector(const char *, char ***);
+static struct ChildRecord *CreateChild(const char *, const char *, SECURITY_ATTRIBUTES *, HANDLE, HANDLE, HANDLE);
 static bool has_redirection(const char *);
 static void StartSockets ();
 static DWORD wait_events(HANDLE event, DWORD timeout);
-#if !defined(__BORLANDC__) && !defined(_WIN32_WCE)
+#if !defined(_WIN32_WCE)
 static int rb_w32_open_osfhandle(long osfhandle, int flags);
 #else
 #define rb_w32_open_osfhandle(osfhandle, flags) _open_osfhandle(osfhandle, flags)
@@ -293,6 +293,7 @@ flock_winnt(VALUE self, int argc, VALUE* argv)
 			  0, LK_LEN, LK_LEN, &o), i);
 	break;
       case LOCK_UN:		/* unlock lock */
+      case LOCK_UN|LOCK_NB:	/* unlock is always non-blocking, I hope */
 	LK_ERR(UnlockFileEx(fh, 0, LK_LEN, LK_LEN, &o), i);
 	break;
       default:            /* unknown */
@@ -320,6 +321,7 @@ flock_win95(VALUE self, int argc, VALUE* argv)
 	LK_ERR(LockFile(fh, 0, 0, LK_LEN, LK_LEN), i);
 	break;
       case LOCK_UN:
+      case LOCK_UN|LOCK_NB:
 	LK_ERR(UnlockFile(fh, 0, 0, LK_LEN, LK_LEN), i);
 	break;
       default:
@@ -363,15 +365,6 @@ NtInitialize(int *argc, char ***argv)
     WORD version;
     int ret;
 
-#ifdef __BORLANDC__
-    _controlfp(0x5, 0x5);
-#endif
-
-    //
-    // subvert cmd.exe's feeble attempt at command line parsing
-    //
-    *argc = make_cmdvector(GetCommandLine(), argv);
-
     //
     // Now set up the correct time stuff
     //
@@ -380,11 +373,6 @@ NtInitialize(int *argc, char ***argv)
 
     // Initialize Winsock
     StartSockets();
-
-#ifdef _WIN32_WCE
-    // free commandline buffer
-    wce_FreeCommandLine();
-#endif
 }
 
 char *getlogin()
@@ -863,15 +851,17 @@ char **argv;
 }
 
 static struct ChildRecord *
-CreateChild(char *cmd, char *prog, SECURITY_ATTRIBUTES *psa, HANDLE hInput, HANDLE hOutput, HANDLE hError)
+CreateChild(const char *cmd, const char *prog, SECURITY_ATTRIBUTES *psa,
+	    HANDLE hInput, HANDLE hOutput, HANDLE hError)
 {
     BOOL fRet;
     DWORD  dwCreationFlags;
     STARTUPINFO aStartupInfo;
     PROCESS_INFORMATION aProcessInformation;
     SECURITY_ATTRIBUTES sa;
-    char *shell;
+    const char *shell;
     struct ChildRecord *child;
+    char *p = NULL;
 
     if (!cmd && !prog) {
 	errno = EFAULT;
@@ -919,31 +909,77 @@ CreateChild(char *cmd, char *prog, SECURITY_ATTRIBUTES *psa, HANDLE hInput, HAND
     dwCreationFlags = (NORMAL_PRIORITY_CLASS);
 
     if (prog) {
-	shell = prog;
+	if (!(p = dln_find_exe(prog, NULL))) {
+	    shell = prog;
+	}
     }
     else {
 	int redir = -1;
+	int len = 0;
+	while (ISSPACE(*cmd)) cmd++;
+	for (prog = cmd; *prog; prog = CharNext(prog)) {
+	    if (ISSPACE(*prog)) {
+		len = prog - cmd;
+		do ++prog; while (ISSPACE(*prog));
+		if (!*prog--) break;
+	    }
+	    else {
+		len = 0;
+	    }
+	}
+	if (!len) len = strlen(cmd);
 	if ((shell = getenv("RUBYSHELL")) && (redir = has_redirection(cmd))) {
-	    char *tmp = ALLOCA_N(char, strlen(shell) + strlen(cmd) +
-				 sizeof (" -c ") + 2);
-	    sprintf(tmp, "%s -c \"%s\"", shell, cmd);
+	    char *tmp = ALLOCA_N(char, strlen(shell) + len + sizeof(" -c ") + 2);
+	    sprintf(tmp, "%s -c \"%.*s\"", shell, len, cmd);
 	    cmd = tmp;
 	}
 	else if ((shell = getenv("COMSPEC")) &&
 		 ((redir < 0 ? has_redirection(cmd) : redir) ||
 		  isInternalCmd(cmd, shell))) {
-	    char *tmp = ALLOCA_N(char, strlen(shell) + strlen(cmd) +
-				 sizeof (" /c "));
-	    sprintf(tmp, "%s /c %s", shell, cmd);
+	    char *tmp = ALLOCA_N(char, strlen(shell) + len + sizeof(" /c "));
+	    sprintf(tmp, "%s /c %.*s", shell, len, cmd);
 	    cmd = tmp;
 	}
 	else {
 	    shell = NULL;
+	    prog = cmd;
+	    for (;;) {
+		if (!*prog) {
+		    p = dln_find_exe(cmd, NULL);
+		    break;
+		}
+		if (strchr(".:*?\"/\\", *prog)) {
+		    if (cmd[len]) {
+			char *tmp = ALLOCA_N(char, len + 1);
+			memcpy(tmp, cmd, len);
+			tmp[len] = 0;
+			cmd = tmp;
+		    }
+		    break;
+		}
+		if (ISSPACE(*prog) || strchr("<>|", *prog)) {
+		    len = prog - cmd;
+		    p = ALLOCA_N(char, len + 1);
+		    memcpy(p, cmd, len);
+		    p[len] = 0;
+		    p = dln_find_exe(p, NULL);
+		    break;
+		}
+		prog++;
+	    }
+	}
+    }
+    if (p) {
+	shell = p;
+	while (*p) {
+	    if ((unsigned char)*p == '/')
+		*p = '\\';
+	    p = CharNext(p);
 	}
     }
 
     RUBY_CRITICAL({
-	fRet = CreateProcess(shell, cmd, psa, psa,
+	fRet = CreateProcess(shell, (char *)cmd, psa, psa,
 			     psa->bInheritHandle, dwCreationFlags, NULL, NULL,
 			     &aStartupInfo, &aProcessInformation);
 	errno = map_errno(GetLastError());
@@ -1084,8 +1120,8 @@ skipspace(char *ptr)
     return ptr;
 }
 
-static int 
-make_cmdvector(const char *cmd, char ***vec)
+int 
+rb_w32_cmdvector(const char *cmd, char ***vec)
 {
     int cmdlen, globbing, len, i;
     int elements, strsz, done;
@@ -1599,6 +1635,19 @@ rb_w32_open_osfhandle(long osfhandle, int flags)
 
 #endif
 
+#ifdef __BORLANDC__
+static int
+rb_w32_open_osfhandle(long osfhandle, int flags)
+{
+    int fd = _open_osfhandle(osfhandle, flags);
+    if (fd == -1) {
+	errno = EMFILE;		/* too many open files */
+	_doserrno = 0L;		/* not an OS error */
+    }
+    return fd;
+}
+#endif
+
 #undef getsockopt
 
 static int
@@ -1773,7 +1822,12 @@ rb_w32_fdclr(int fd, fd_set *set)
 int
 rb_w32_fdisset(int fd, fd_set *set)
 {
-       return __WSAFDIsSet(TO_SOCKET(fd), set);
+    int ret;
+    SOCKET s = TO_SOCKET(fd);
+    if (s == (SOCKET)INVALID_HANDLE_VALUE)
+        return 0;
+    RUBY_CRITICAL(ret = __WSAFDIsSet(s, set));
+    return ret;
 }
 
 //
@@ -1970,8 +2024,16 @@ rb_w32_connect(int s, struct sockaddr *addr, int addrlen)
     }
     RUBY_CRITICAL({
 	r = connect(TO_SOCKET(s), addr, addrlen);
-	if (r == SOCKET_ERROR)
-	    errno = map_errno(WSAGetLastError());
+	if (r == SOCKET_ERROR) {
+	    r = WSAGetLastError();
+	    if (r != WSAEWOULDBLOCK) {
+		errno = map_errno(r);
+	    }
+	    else {
+		errno = EINPROGRESS;
+		r = -1;
+	    }
+	}
     });
     return r;
 }
@@ -2334,6 +2396,43 @@ void setprotoent (int stayopen) {}
 
 void setservent (int stayopen) {}
 
+int
+fcntl(int fd, int cmd, ...)
+{
+    SOCKET sock = TO_SOCKET(fd);
+    va_list va;
+    int arg;
+    int ret;
+    u_long ioctlArg;
+
+    if (!is_socket(sock)) {
+	errno = EBADF;
+	return -1;
+    }
+    if (cmd != F_SETFL) {
+	errno = EINVAL;
+	return -1;
+    }
+
+    va_start(va, cmd);
+    arg = va_arg(va, int);
+    va_end(va);
+    if (arg & O_NONBLOCK) {
+	ioctlArg = 1;
+    }
+    else {
+	ioctlArg = 0;
+    }
+    RUBY_CRITICAL({
+	ret = ioctlsocket(sock, FIONBIO, &ioctlArg);
+	if (ret == -1) {
+	    errno = map_errno(WSAGetLastError());
+	}
+    });
+
+    return ret;
+}
+
 #ifndef WNOHANG
 #define WNOHANG -1
 #endif
@@ -2687,7 +2786,7 @@ isUNCRoot(const char *path)
 {
     if (path[0] == '\\' && path[1] == '\\') {
 	const char *p;
-	for (p = path + 3; *p; p = CharNext(p)) {
+	for (p = path + 2; *p; p = CharNext(p)) {
 	    if (*p == '\\')
 		break;
 	}
@@ -2696,7 +2795,7 @@ isUNCRoot(const char *path)
 		if (*p == '\\')
 		    break;
 	    }
-	    if (!p[0] || !p[1] || (p[0] == '\\' && p[1] == '.'))
+	    if (!p[0] || !p[1] || (p[1] == '.' && !p[2]))
 		return 1;
 	}
     }
@@ -2865,7 +2964,7 @@ static void rb_w32_call_handler(struct handler_arg_t* h)
 	rb_jump_tag(status);
     }
     h->finished = 1;
-    Sleep(INFINITE);		/* safe on Win95? */
+    yield_until(0);
 }
 
 static struct handler_arg_t* setup_handler(struct handler_arg_t *harg,
@@ -3146,11 +3245,13 @@ void rb_w32_free_environ(char **env)
     free(env);
 }
 
+#undef getpid
 pid_t rb_w32_getpid(void)
 {
     pid_t pid;
 
-    pid = _getpid();
+    pid = getpid();
+
     if (IsWin95()) pid = -pid;
 
     return pid;
@@ -3185,6 +3286,8 @@ rb_w32_close(int fd)
 	UnlockFile((HANDLE)sock, 0, 0, LK_LEN, LK_LEN);
 	return _close(fd);
     }
+    _set_osfhnd(fd, (SOCKET)INVALID_HANDLE_VALUE);
+    _close(fd);
     if (closesocket(sock) == SOCKET_ERROR) {
 	errno = map_errno(WSAGetLastError());
 	return -1;
@@ -3277,3 +3380,110 @@ rb_w32_snprintf(char *buf, size_t size, const char *format, ...)
     va_end(va);
     return ret;
 }
+
+#undef mkdir
+#undef rmdir
+int
+rb_w32_mkdir(const char *path, int mode)
+{
+    int ret = -1;
+    RUBY_CRITICAL(do {
+	if (mkdir(path) == -1)
+	    break;
+	if (chmod(path, mode) == -1) {
+	    int save_errno = errno;
+	    rmdir(path);
+	    errno = save_errno;
+	    break;
+	}
+	ret = 0;
+    } while (0));
+    return ret;
+}
+
+int
+rb_w32_rmdir(const char *path)
+{
+    DWORD attr;
+    int ret;
+    RUBY_CRITICAL({
+	attr = GetFileAttributes(path);
+	if (attr != (DWORD)-1 && (attr & FILE_ATTRIBUTE_READONLY)) {
+	    attr &= ~FILE_ATTRIBUTE_READONLY;
+	    SetFileAttributes(path, attr);
+	}
+	ret = rmdir(path);
+	if (ret < 0 && attr != (DWORD)-1) {
+	    SetFileAttributes(path, attr);
+	}
+    });
+    return ret;
+}
+
+#if !defined(__BORLANDC__) && !defined(_WIN32_WCE)
+int
+rb_w32_isatty(int fd)
+{
+    if (!(_osfile(fd) & FOPEN)) {
+	errno = EBADF;
+	return 0;
+    }
+    if (!(_osfile(fd) & FDEV)) {
+	errno = ENOTTY;
+	return 0;
+    }
+    return 1;
+}
+#endif
+
+//
+// Fix bcc32's stdio bug
+//
+
+#ifdef __BORLANDC__
+static int
+too_many_files()
+{
+    FILE *f;
+    for (f = _streams; f < _streams + _nfile; f++) {
+	if (f->fd < 0) return 0;
+    }
+    return 1;
+}
+
+#undef fopen
+FILE *
+rb_w32_fopen(const char *path, const char *mode)
+{
+    FILE *f = (errno = 0, fopen(path, mode));
+    if (f == NULL && errno == 0) {
+	if (too_many_files())
+	    errno = EMFILE;
+    }
+    return f;
+}
+
+FILE *
+rb_w32_fdopen(int handle, const char *type)
+{
+    FILE *f = (errno = 0, _fdopen(handle, (char *)type));
+    if (f == NULL && errno == 0) {
+	if (handle < 0)
+	    errno = EBADF;
+	else if (too_many_files())
+	    errno = EMFILE;
+    }
+    return f;
+}
+
+FILE *
+rb_w32_fsopen(const char *path, const char *mode, int shflags)
+{
+    FILE *f = (errno = 0, _fsopen(path, mode, shflags));
+    if (f == NULL && errno == 0) {
+	if (too_many_files())
+	    errno = EMFILE;
+    }
+    return f;
+}
+#endif
